
from vsg import parser


class miscellaneous_operator(parser.keyword):
    '''
    unique_id = miscellaneous_operator : miscellaneous_operator
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class double_star(miscellaneous_operator):
    '''
    unique_id = miscellaneous_operator : double_star
    '''

    def __init__(self, sString='**'):
        miscellaneous_operator.__init__(self, '**')


class abs_operator(miscellaneous_operator):
    '''
    unique_id = miscellaneous_operator : abs_operator
    '''

    def __init__(self, sString):
        miscellaneous_operator.__init__(self, sString)


class not_operator(miscellaneous_operator):
    '''
    unique_id = miscellaneous_operator : not_operator
    '''

    def __init__(self, sString):
        miscellaneous_operator.__init__(self, sString)




from vsg.rules import token_indent as Rule

from vsg import token

lTokens = []
lTokens.append(token.entity_declaration.begin_keyword)


class rule_300(Rule):
    '''
    This rule checks the indent of the **begin** keyword.

    **Violation**

    .. code-block:: vhdl

       entity FIFO is

            begin

    **Fix**

    .. code-block:: vhdl

       entity FIFO is

       begin
    '''

    def __init__(self):
        Rule.__init__(self, 'entity', '300', lTokens)




from vsg import parser
from vsg import violation

from vsg.rules.utils import token_is_whitespace, token_is_comment

from vsg.rule_group import blank_line


class previous_line(blank_line.Rule):
    '''
    Checks for a blank line above a line starting with a given token

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    token: token object type list
       token object that a blank line above should appear
    '''

    def __init__(self, name, identifier, lTokens, lAllowTokens=None):
        blank_line.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = 'Insert blank line above'
        self.lTokens = lTokens
        self.lHierarchyLimits = None
        self.style = 'require_blank_line'
        self.configuration.append('style')
        if lAllowTokens is None:
            self.lAllowTokens = []
        else:
            self.lAllowTokens = lAllowTokens
        self.configuration_documentation_link = 'configuring_previous_line_rules_link'

    def _get_tokens_of_interest(self, oFile):
        bIncludeComments = _include_comments(self.style)

        if self.style == 'require_comment':
            if self.lHierarchyLimits is None:
                lFirst = oFile.get_line_above_line_starting_with_token(self.lTokens, False)
                lSecond = oFile.get_line_above_line_starting_with_token(self.lTokens, True)
                return zip(lFirst, lSecond)
            else:
                lFirst = oFile.get_line_above_line_starting_with_token_with_hierarchy(self.lTokens, self.lHierarchyLimits, False)
                lSecond = oFile.get_line_above_line_starting_with_token_with_hierarchy(self.lTokens, self.lHierarchyLimits, True)
                return zip(lFirst, lSecond)

        elif self.style == 'no_blank_line':
            return oFile.get_blank_lines_above_line_starting_with_token(self.lTokens)
        else:

            if self.lHierarchyLimits is None:
                return oFile.get_line_above_line_starting_with_token(self.lTokens, bIncludeComments)
            else:
                return oFile.get_line_above_line_starting_with_token_with_hierarchy(self.lTokens, self.lHierarchyLimits, bIncludeComments)

    def _set_allow_tokens(self):
        return

    def _analyze(self, lToi):
        self._set_allow_tokens()
        if self.style == 'no_blank_line':
            _analyze_no_blank_line(self, lToi, self.lAllowTokens)
        elif self.style == 'require_blank_line':
            _analyze_require_blank_line(self, lToi, self.lAllowTokens)
        elif self.style == 'no_code':
            _analyze_no_code(self, lToi, self.lAllowTokens)
        elif self.style == 'allow_comment':
            _analyze_no_code(self, lToi, self.lAllowTokens)
        elif self.style == 'require_comment':
            _analyze_require_comment(self, lToi, self.lAllowTokens)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        if dAction['action'] == 'Insert':
            lTokens.append(parser.carriage_return())
            lTokens.append(parser.blank_line())
            oViolation.set_tokens(lTokens)
        elif dAction['action'] == 'Remove':
            oViolation.set_tokens([])


def _include_comments(sMethod):
    if sMethod == 'require_blank_line':
        return False
    elif sMethod == 'no_code':
        return False
    elif sMethod == 'allow_comment':
        return True
    elif sMethod == 'require_comment':
        return True
    return None


def _analyze_no_blank_line(self, lToi, lAllowTokens):
        for oToi in lToi:
            sSolution = 'Remove blank lines'
            oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
            dAction = {}
            dAction['action'] = 'Remove'
            oViolation.set_action(dAction)
            self.add_violation(oViolation)


def _analyze_require_blank_line(self, lToi, lAllowTokens):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if _is_allowed_token(lAllowTokens, lTokens):
                continue
            if len(lTokens) == 1:
                if isinstance(lTokens[0], parser.blank_line):
                    continue
            oViolation = violation.New(oToi.get_line_number(), oToi, self.solution)
            dAction = {}
            dAction['action'] = 'Insert'
            oViolation.set_action(dAction)
            self.add_violation(oViolation)


def _analyze_no_code(self, lToi, lAllowTokens):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if _is_allowed_token(lAllowTokens, lTokens):
                continue
            if len(lTokens) == 1:
                if isinstance(lTokens[0], parser.blank_line) or token_is_comment(lTokens[0]):
                    continue
            elif len(lTokens) == 2:
                if token_is_whitespace(lTokens[0]) and token_is_comment(lTokens[1]):
                    continue
            oViolation = violation.New(oToi.get_line_number(), oToi, self.solution)
            dAction = {}
            dAction['action'] = 'Insert'
            oViolation.set_action(dAction)
            self.add_violation(oViolation)


def _analyze_require_comment(self, lToi, lAllowTokens):
        for oFirstToi, oToi in list(lToi):
            lTokens = oFirstToi.get_tokens()
            if _is_allowed_token(lAllowTokens, lTokens):
                continue
            #Check if first line is a comment
            if not _comment_starts_line(lTokens):
                sSolution = 'Comment required above this line.'
                oViolation = violation.New(oFirstToi.get_line_number(), oFirstToi, sSolution)
                dAction = {}
                dAction['action'] = 'Skip'
                oViolation.set_action(dAction)
                self.add_violation(oViolation)
                continue

            lTokens = oToi.get_tokens()
            if _is_allowed_token(lAllowTokens, lTokens):
                continue
            if len(lTokens) == 1:
                if isinstance(lTokens[0], parser.blank_line):
                    continue
            elif len(lTokens) == 2:
                if token_is_whitespace(lTokens[0]) and token_is_comment(lTokens[1]) and self.allow_comment:
                    continue
            oViolation = violation.New(oToi.get_line_number(), oToi, self.solution)
            dAction = {}
            dAction['action'] = 'Insert'
            oViolation.set_action(dAction)
            self.add_violation(oViolation)


def _is_allowed_token(lAllowTokens, lTokens):
    bSkip = False
    for oAllowToken in lAllowTokens:
        for oToken in lTokens:
            if isinstance(oToken, oAllowToken):
                bSkip = True
                break
        if bSkip:
           break
    if bSkip:
        return True
    return False


def _comment_starts_line(lTokens):
    if len(lTokens) == 1:
      if token_is_comment(lTokens[0]):
          return True
    elif len(lTokens) == 2:
      if token_is_whitespace(lTokens[0]) and token_is_comment(lTokens[1]):
          return True
    return False




from vsg import parser


class rising_edge(parser.function):
    '''
    unique_id = ieee_std_logic_1164_function : rising_edge
    '''

    def __init__(self, sString):
        parser.function.__init__(self, sString)


class falling_edge(parser.function):
    '''
    unique_id = ieee_std_logic_1164_function : falling_edge
    '''

    def __init__(self, sString):
        parser.function.__init__(self, sString)




from vsg import token

from vsg.rules import blank_line_below_line_ending_with_token

lTokens = []
lTokens.append(token.incomplete_type_declaration.semicolon)
lTokens.append(token.full_type_declaration.semicolon)


class rule_011(blank_line_below_line_ending_with_token):
    '''
    This rule checks for a blank line below the **type** declaration.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       type state_machine is (idle, write, read, done);
       signal sm : state_machine;

    **Fix**

    .. code-block:: vhdl

       type state_machine is (idle, write, read, done);

       signal sm : state_machine;
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'type', '011', lTokens)




from vsg.token import configuration_declaration as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import configuration_declarative_part
from vsg.vhdlFile.classify import block_configuration


def detect(iToken, lObjects):
    '''
    configuration_declaration ::=
      configuration identifier of *entity*_name is
        configuration_declarative_part
        { verification_unit_binding_indication ; }
      block_configuration
      end [ configuration ] [ *configuration*_simple_name ] ;
    '''

    if utils.is_next_token('configuration', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = classify_opening_declaration(iToken, lObjects)

    iCurrent = configuration_declarative_part.detect(iCurrent, lObjects)

    iCurrent = block_configuration.detect(iCurrent, lObjects)

    iCurrent = classify_closing_declaration(iToken, lObjects)

    return iCurrent


def classify_opening_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('configuration', token.configuration_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token(token.identifier, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('of', token.of_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token(token.entity_name, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)

    return iCurrent


def classify_closing_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if('configuration', token.end_configuration_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.configuration_simple_name, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import selected_force_assignment
from vsg.vhdlFile.classify import selected_waveform_assignment


def detect(iToken, lObjects):
    '''
    selected_signal_assignment ::=
        selected_waveform_assignment
      | selected_force_assignment
    '''

    if utils.find_in_range('<=', iToken, ';', lObjects):
        if utils.find_in_next_n_tokens('with', 3, iToken, lObjects):
            return True
        if utils.find_in_next_n_tokens('if', 3, iToken, lObjects):
            return True
    return False


def classify(iToken, lObjects):
    iCurrent = selected_waveform_assignment.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = selected_force_assignment.detect(iToken, lObjects)

    return iCurrent




from .rule_001 import rule_001
from .rule_002 import rule_002
from .rule_003 import rule_003
from .rule_004 import rule_004
from .rule_005 import rule_005
from .rule_006 import rule_006
from .rule_007 import rule_007
from .rule_008 import rule_008
from .rule_009 import rule_009
from .rule_010 import rule_010

from .rule_013 import rule_013
from .rule_014 import rule_014

from .rule_016 import rule_016
from .rule_017 import rule_017
from .rule_018 import rule_018
from .rule_019 import rule_019
from .rule_020 import rule_020
from .rule_021 import rule_021

from .rule_600 import rule_600




from vsg import token

from vsg.rules.whitespace_before_token import Rule

lTokens = []
lTokens.append(token.constant_declaration.colon)


class rule_006(Rule):
    '''
    This rule checks for at least a single space before the colon.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       constant size: integer := 1;
       constant width     : integer := 32;

    **Fix**

    .. code-block:: vhdl

       constant size : integer := 1;
       constant width     : integer := 32;
    '''
    def __init__(self):
        Rule.__init__(self, 'constant', '006', lTokens)
        self.number_of_spaces = '>=1'




from vsg import parser
from vsg import violation

from vsg.rule_group import structure


class remove_comments_from_end_of_lines_bounded_by_tokens(structure.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of parser object types
       object types to check the prefix

    lPrefixes : string list
       acceptable prefixes
    '''

    def __init__(self, name, identifier, oStart, oEnd):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.oStart = oStart
        self.oEnd = oEnd
        self.configuration_documentation_link = None

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(self.oStart, self.oEnd)

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iLine = oToi.get_line_number()
            for iToken, oToken in enumerate(lTokens):
                if isinstance(oToken, parser.carriage_return):
                    iLine += 1
                if isinstance(oToken, parser.comment):
                    if isinstance(lTokens[iToken + 1], parser.carriage_return):
                        if isinstance(lTokens[iToken - 1], parser.carriage_return) or \
                           isinstance(lTokens[iToken - 2], parser.carriage_return):
                            continue
                        else:
                            if isinstance(lTokens[iToken - 1], parser.whitespace):
                                oNewToi = oToi.extract_tokens(iToken - 1, iToken)
                            else:
                                oNewToi = oToi.extract_tokens(iToken, iToken)
                            oViolation = violation.New(oNewToi.get_line_number(), oNewToi, self.solution)
                            self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        for oViolation in self.violations:
            oViolation.set_tokens([])




from vsg import parser

from vsg.vhdlFile.extract import tokens
from vsg.vhdlFile.extract import utils


def get_blank_lines_below_line_ending_with_several_possible_tokens(lTokens, lAllTokens, oTokenMap):

    lReturn = []

    lIndexes = utils.get_indexes_of_tokens_between(lTokens[0], lTokens[1:], oTokenMap)

    lCarriageReturns = oTokenMap.get_token_indexes(parser.carriage_return)
    lBlankLines = oTokenMap.get_token_indexes(parser.blank_line)

    for iIndex in lIndexes:
        if not utils.is_token_at_end_of_line(iIndex, oTokenMap):
            continue
        iLine = oTokenMap.get_line_number_of_index(iIndex)
        iStart = lCarriageReturns[iLine - 1] + 1
        for i in range(iLine - 1, len(lCarriageReturns)):
            iCarriageReturnIndex = lCarriageReturns[i]
            if not iCarriageReturnIndex + 1 in lBlankLines:
                iEnd = lCarriageReturns[i] + 1
                lTemp = lAllTokens[iStart:iEnd]
                if len(lTemp) > 0:
                    lReturn.append(tokens.New(iStart, iLine, lTemp))
                break

    return lReturn






from vsg.token import subprogram_body as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import subprogram_declarative_part
from vsg.vhdlFile.classify import subprogram_kind
from vsg.vhdlFile.classify import subprogram_statement_part


def detect(iToken, lObjects):
    '''
    subprogram_body ::=
        subprogram_specification is
            subprogram_declarative_part
        begin
            subprogram_statement_part
        end [ subprogram_kind ] [ designator ] ;
    '''

    if utils.is_next_token('is', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iToken, lObjects)

    iCurrent = subprogram_declarative_part.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('begin', token.begin_keyword, iCurrent, lObjects)

    iCurrent = subprogram_statement_part.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iCurrent, lObjects)

    if subprogram_kind.detect(iCurrent, lObjects):
        iCurrent = subprogram_kind.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_if_not(';', token.designator, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.signal_declaration.signal_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of signal declarations.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is

       signal wr_en : std_logic;
            signal rd_en : std_logic;

       begin

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is

         signal wr_en : std_logic;
         signal rd_en : std_logic;

       begin
    '''

    def __init__(self):
        token_indent.__init__(self, 'signal', '001', lTokens)




from vsg.rules import insert_token_left_of_token_if_it_does_not_exist_between_tokens_using_value_from_token as Rule

from vsg.token import loop_statement as token

oInsertToken = token.end_loop_label

oAnchorToken = token.semicolon

oLeftToken = token.end_loop_keyword
oRightToken = token.semicolon

oValueToken = token.loop_label


class rule_007(Rule):
    '''
    This rule checks the **end loop_statement** line has a label.
    The closing label will be added if the opening loop_statement label exists.

    |configuring_optional_items_link|

    **Violation**

    .. code-block:: vhdl

       end loop;

    **Fix**

    .. code-block:: vhdl

       end loop LOOP_LABEL;
    '''

    def __init__(self):
        Rule.__init__(self, 'loop_statement', '007', oInsertToken, oAnchorToken, oLeftToken, oRightToken, oValueToken)
        self.solution = 'a label for the "end loop".'
        self.subphase = 2
        self.disable = True




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.signal_assignment_statement.label)
lTokens.append(token.simple_force_assignment.target)
lTokens.append(token.simple_waveform_assignment.target)
lTokens.append(token.conditional_waveform_assignment.target)


class rule_001(token_indent):
    '''
    This rule checks the indent of sequential statements.

    **Violation**

    .. code-block:: vhdl

       begin

           wr_en <= '1';
       rd_en <= '0';

    **Fix**

    .. code-block:: vhdl

       begin

         wr_en <= '1';
         rd_en <= '0';
    '''

    def __init__(self):
        token_indent.__init__(self, 'sequential', '001', lTokens)




from vsg.vhdlFile.classify import alias_declaration
from vsg.vhdlFile.classify import attribute_declaration
from vsg.vhdlFile.classify import attribute_specification
from vsg.vhdlFile.classify import component_declaration
from vsg.vhdlFile.classify import constant_declaration
from vsg.vhdlFile.classify import file_declaration
from vsg.vhdlFile.classify import package_body
from vsg.vhdlFile.classify import package_declaration
from vsg.vhdlFile.classify import package_instantiation_declaration
from vsg.vhdlFile.classify import signal_declaration
from vsg.vhdlFile.classify import subprogram_body
from vsg.vhdlFile.classify import subprogram_declaration
from vsg.vhdlFile.classify import subprogram_instantiation_declaration
from vsg.vhdlFile.classify import subtype_declaration
from vsg.vhdlFile.classify import type_declaration
from vsg.vhdlFile.classify import use_clause
from vsg.vhdlFile.classify import variable_declaration


def detect(iToken, lObjects):
    '''
    entity_declarative_item ::=
        subprogram_declaration
      | subprogram_body
      | subprogram_instantiation_declaration
      | package_declaration
      | package_body
      | package_instantiation_declaration
      | type_declaration
      | subtype_declaration
      | constant_declaration
      | signal_declaration
      | *shared*_variable_declaration
      | file_declaration
      | alias_declaration
      | attribute_declaration
      | attribute_specification
      | disconnection_specification
      | use_clause
      | group_template_declaration
      | group_declaration
      | PSL_Property_Declaration
      | PSL_Sequence_Declaration
      | PSL_Clock_Declaration
    '''

    iCurrent = subprogram_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        iCurrent = subprogram_body.detect(iCurrent, lObjects)
        return iCurrent

    iCurrent = subprogram_instantiation_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = package_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = package_body.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = package_instantiation_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = type_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = subtype_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = constant_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = signal_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = variable_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = file_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = alias_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = component_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = attribute_declaration.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = attribute_specification.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent



    iCurrent = use_clause.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent


    return iToken




from vsg import parser
from vsg.vhdlFile import utils

from vsg.vhdlFile.extract import get_tokens_in_architecture_declarative_part
from vsg.vhdlFile.extract import get_tokens_in_package_declarative_part
from vsg.vhdlFile.extract import get_tokens_in_package_body_declarative_part
from vsg.vhdlFile.extract import get_tokens_in_subprogram_declarative_part
from vsg.vhdlFile.extract import get_tokens_in_process_declarative_part
from vsg.vhdlFile.extract import get_tokens_in_entity_declarative_part
from vsg.vhdlFile.extract import get_tokens_in_block_declarative_part
from vsg.vhdlFile.extract import get_tokens_in_protected_type_body_declarative_part


def get_tokens_in_declarative_parts(lAllTokens, oTokenMap):

    lProtectedType = get_tokens_in_protected_type_body_declarative_part.extract(lAllTokens, oTokenMap)
    lArchitecture = get_tokens_in_architecture_declarative_part.extract(lAllTokens, oTokenMap)
    lPackageBody = get_tokens_in_package_body_declarative_part.extract(lAllTokens, oTokenMap)
    lSubprogram = get_tokens_in_subprogram_declarative_part.extract(lAllTokens, oTokenMap)
    lPackage = get_tokens_in_package_declarative_part.extract(lAllTokens, oTokenMap)
    lProcess = get_tokens_in_process_declarative_part.extract(lAllTokens, oTokenMap)
    lEntity = get_tokens_in_entity_declarative_part.extract(lAllTokens, oTokenMap)
    lBlock = get_tokens_in_block_declarative_part.extract(lAllTokens, oTokenMap)

    lReturn = utils.combine_two_token_class_lists(lArchitecture, lProtectedType)
    lReturn = utils.combine_two_token_class_lists(lReturn, lPackage)
    lReturn = utils.combine_two_token_class_lists(lReturn, lPackageBody)
    lReturn = utils.combine_two_token_class_lists(lReturn, lSubprogram)
    lReturn = utils.combine_two_token_class_lists(lReturn, lProcess)
    lReturn = utils.combine_two_token_class_lists(lReturn, lEntity)
    lReturn = utils.combine_two_token_class_lists(lReturn, lBlock)

    return lReturn





from vsg import parser
from vsg import violation

from vsg.rule_group import structure
from vsg.rules import utils as rules_utils

from vsg.vhdlFile import utils


class split_line_at_token_if_on_same_line_as_token_if_token_pair_are_not_on_the_same_line(structure.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of parser object types
       object type to split a line at

    oStart : token type
       The start of the range

    oEnd : token type
       The end of the range
    '''

    def __init__(self, name, identifier, oToken, oSameLineToken, lTokenPair):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = None
        self.phase = 1
        self.oToken = oToken
        self.oSameLineToken = oSameLineToken
        self.lTokenPair = lTokenPair
        self.configuration_documentation_link = None

    def _get_tokens_of_interest(self, oFile):
        lToi = oFile.get_tokens_bounded_by(self.lTokenPair[0], self.lTokenPair[1])
        lToi = extract_multi_line_statements(lToi)
        return lToi

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iLine = oToi.get_line_number()
            for iToken, oToken in enumerate(lTokens):
                if isinstance(oToken, parser.carriage_return):
                    iLine += 1
                if isinstance(oToken, self.oSameLineToken):
                    if utils.are_next_consecutive_token_types([parser.whitespace, self.oToken], iToken + 1, lTokens) or \
                       utils.are_next_consecutive_token_types([self.oToken], iToken + 1, lTokens):
                        sSolution = self.solution
                        oViolation = violation.New(iLine, oToi, sSolution)
                        dAction = {}
                        dAction['insert_index'] = iToken + 1
                        oViolation.set_action(dAction)
                        self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        rules_utils.insert_carriage_return(lTokens, dAction['insert_index'])
        oViolation.set_tokens(lTokens)


def extract_multi_line_statements(lToi):
    lReturn = []
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        for oToken in lTokens:
            if isinstance(oToken, parser.carriage_return):
                lReturn.append(oToi)
                break
    return lReturn




from vsg import parser
from vsg.vhdlFile import utils


def build_solution(sIndent):
    sSolution = 'Indent with '
    if '\t' in sIndent and ' ' in sIndent:
        sSolution += str(sIndent.count('\t')) + ' tab(s) followed by ' + str(sIndent.count(' ')) + ' space(s)'
    elif '\t' in sIndent:
        sSolution += str(sIndent.count('\t')) + ' tab(s)'
    elif ' ' in sIndent:
        sSolution += str(sIndent.count(' ')) + ' space(s)'
    return sSolution


def convert_expected_indent_to_smart_tab(dExpectedIndent, indentSize, iFirstLineIndent):
    iFirstLine = get_first_line(dExpectedIndent)
    iLastLine = get_last_line(dExpectedIndent)
    for iLine in range(iFirstLine + 1, iLastLine + 1):
        dExpectedIndent[iLine] = '\t' + dExpectedIndent[iLine][iFirstLineIndent:]


def get_first_line(dActualIndent):
    lLines = list(dActualIndent.keys())
    lLines.sort()
    iLine = lLines[0]
    return iLine


def get_first_line_info(iLine, oFile):
    lTemp = oFile.get_tokens_from_line(iLine)
    iIndent = len(lTemp.get_tokens()[0].get_value())
    return iLine, iIndent


def get_last_line(dActualIndent):
    lLines = list(dActualIndent.keys())
    lLines.sort()
    iLine = lLines[-1]
    return iLine


def set_indent(iToken, lTokens):
    iReturn = 0
    if isinstance(lTokens[iToken + 1], parser.whitespace):
        iReturn = lTokens[iToken + 1].get_value()
    else:
        iReturn = ''
    return iReturn


def starts_with_paren(lTokens):
    iToken = utils.find_next_non_whitespace_token(1, lTokens)
    try:
        if isinstance(lTokens[iToken], parser.open_parenthesis):
            return True
    except IndexError:
        if isinstance(lTokens[0], parser.open_parenthesis):
            return True
    return False


def update_column_width(self, oToken):
    sToken = oToken.get_value()
    if isinstance(oToken, parser.whitespace):
        iTabs = sToken.count('\t')
        iLength = len(sToken) + (iTabs * self.indentSize) - iTabs
        return iLength

    return len(oToken.get_value())




from vsg.vhdlFile.extract import tokens
from vsg.vhdlFile.extract import utils


def get_token_and_n_tokens_before_it(lTokens, iTokens, lAllTokens, oTokenMap):
    lReturn = []

    lIndexes = utils.get_indexes_of_token_list(lTokens, oTokenMap)

    for iIndex in lIndexes:
        iLine = oTokenMap.get_line_number_of_index(iIndex)
        iStart = iIndex - iTokens
        if iStart >= 0:
            lReturn.append(tokens.New(iStart, iLine, lAllTokens[iStart:iIndex + 1]))
    return lReturn




from vsg import parser


class attribute_keyword(parser.keyword):
    '''
    unique_id = attribute_specification : attribute_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class attribute_designator(parser.designator):
    '''
    unique_id = attribute_specification : attribute_designator
    '''

    def __init__(self, sString):
        parser.designator.__init__(self, sString)


class of_keyword(parser.keyword):
    '''
    unique_id = attribute_specification : of_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class is_keyword(parser.keyword):
    '''
    unique_id = attribute_specification : is_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = attribute_specification : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.process_statement.process_label)
lTokens.append(token.process_statement.postponed_keyword)
lTokens.append(token.process_statement.process_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of the process declaration.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is

       begin

       proc_a : process (rd_en, wr_en, data_in, data_out,

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is

       begin

         proc_a : process (rd_en, wr_en, data_in, data_out,
    '''

    def __init__(self):
        token_indent.__init__(self, 'process', '001', lTokens)




from vsg.rules import move_token_next_to_another_token

from vsg.token import generic_clause as token


class rule_018(move_token_next_to_another_token):
    '''
    This rule checks the **generic** keyword is on the same line as the (.

    **Violation**

    .. code-block:: vhdl

      generic
       (

    **Fix**

    .. code-block:: vhdl

      generic (
    '''

    def __init__(self):
        move_token_next_to_another_token.__init__(self, 'generic', '018', token.generic_keyword, token.open_parenthesis)
        self.solution = 'Move the ( to the same line as the *generic* keyword.'




from vsg import parser
from vsg import token

from vsg import tokens

from vsg.token import adding_operator
from vsg.token import aggregate
from vsg.token import direction
from vsg.token import exponent
from vsg.token import logical_operator
from vsg.token import miscellaneous_operator
from vsg.token import multiplying_operator
from vsg.token import relational_operator
from vsg.token import resolution_indication
from vsg.token import sign
from vsg.token import subtype_indication
from vsg.token import type_mark
from vsg.token import unary_logical_operator
from vsg.token import choices

from vsg.token.ieee.std_logic_1164 import types
from vsg.token.ieee.std_logic_1164 import function

from vsg.vhdlFile import extract
from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import blank
from vsg.vhdlFile.classify import comment
from vsg.vhdlFile.classify import design_file
from vsg.vhdlFile.classify import whitespace
from vsg.vhdlFile.classify import preprocessor
from vsg.vhdlFile.classify import pragma

from vsg.vhdlFile.indent.set_token_indent import set_token_indent

from vsg.token_map import process_tokens

from vsg.vhdlFile import code_tags


class vhdlFile():
    '''
    Holds contents of a VHDL file.
    When a vhdlFile object is created, the contents of the file must be passed to it.
    A line object is created for each line read in.
    Then the line object attributes are updated.

    Parameters:

       filecontent: (list)

    Returns:

       fileobject
    '''
    def __init__(self, filecontent, sFilename=None, eError=None):
        self.filecontent = filecontent
        self.hasArchitecture = False
        self.hasEntity = False
        self.lAllObjects = []
        self.filename = sFilename
        self.dIndentMap = None
        self.lOpenPragmas = ['--vhdl_comp_off']
        self.lClosePragmas = ['--vhdl_comp_on']
        self.dVars = {}
        self.dVars['pragma'] = False
        self.eError = eError
        self._processFile()

    def _processFile(self):

        oOptions = options()
        self.lAllObjects = []
        for sLine in self.filecontent:
            lTokens = tokens.create(sLine.rstrip('\n').rstrip('\r'))
            lObjects = []
            for sToken in lTokens:
                lObjects.append(parser.item(sToken))

            blank.classify(lObjects, oOptions)
            whitespace.classify(lTokens, lObjects)
            comment.classify(lTokens, lObjects, oOptions)
            preprocessor.classify(lTokens, lObjects)
            pragma.classify(lTokens, lObjects, self.lOpenPragmas, self.lClosePragmas, self.dVars)

            self.lAllObjects.extend(lObjects)
            self.lAllObjects.append(parser.carriage_return())

        try:
            self.lAllObjects[0].set_filename(self.filename)
        except IndexError:
            pass

        design_file.tokenize(self.lAllObjects)
        post_token_assignments(self.lAllObjects)
        self.lAllObjects = combine_use_clause_selected_name(self.lAllObjects)

        set_token_hierarchy_value(self.lAllObjects)
        set_aggregate_tokens(self.lAllObjects)
        set_code_tags(self.lAllObjects)
        self.oTokenMap = process_tokens(self.lAllObjects)

    def update(self, lUpdates):

        if len(lUpdates) == 0:
            return
        bUpdateMap = True
        for oUpdate in lUpdates[::-1]:
            iStart = oUpdate.oTokens.iStartIndex
            lTokens = oUpdate.get_tokens()
            iEnd = oUpdate.oTokens.iEndIndex
            lMyTokens = remove_beginning_of_file_tokens(lTokens)
            self.lAllObjects[iStart:iEnd] = lMyTokens
        if bUpdateMap:
            self.oTokenMap = process_tokens(self.lAllObjects)

    def update_token_map(self):
        self.oTokenMap = process_tokens(self.lAllObjects)

    def set_indent_map(self, dIndentMap):
        self.dIndentMap = dIndentMap
        set_token_indent(self.dIndentMap, self.lAllObjects)

    def get_indent_map(self):
        return self.dIndentMap

    def get_object_lines(self):
        lReturn = []
        lReturn.append('')
        for lLine in split_on_carriage_return(self.lAllObjects):
            lReturn.append(lLine)
        return lReturn

    def get_lines(self):
        lReturn = []
        lReturn.append('')
        for lLine in split_on_carriage_return(self.lAllObjects):
            lReturn.append(utils.convert_token_list_to_string(lLine))
        return lReturn

    def get_line_count(self):
        return utils.count_carriage_returns(self.lAllObjects)

    def get_line_count_between_tokens(self, oStart, oEnd):
        return extract.get_line_count_between_tokens(oStart, oEnd, self.lAllObjects, self.oTokenMap)

    def fix_blank_lines(self):
        self.lAllObjects = utils.fix_blank_lines(self.lAllObjects)

    def fix_trailing_whitespace(self):
        self.lAllObjects = utils.fix_trailing_whitespace(self.lAllObjects)

    def set_token_indent(self):
        set_token_indent(self.dIndentMap, self.lAllObjects)

    def get_line_preceeding_line(self, iLine, iNumLines=1):
        return extract.get_line_preceeding_line(iLine, self.lAllObjects, iNumLines, self.oTokenMap)

    def get_line_succeeding_line(self, iLine, iNumLines=1):
        return extract.get_line_succeeding_line(iLine, self.lAllObjects, iNumLines, self.oTokenMap)

    def get_all_tokens(self):
        return extract.get_all_tokens(self.lAllObjects)

    def get_sequence_of_tokens_matching(self, lTokens, bIgnoreIfLineStart=False):
        return extract.get_sequence_of_tokens_matching(lTokens, self.lAllObjects, self.oTokenMap, bIgnoreIfLineStart)

    def get_sequence_of_tokens_matching_bounded_by_tokens(self, lTokens, oStart, oEnd):
        return extract.get_sequence_of_tokens_matching_bounded_by_tokens(lTokens, oStart, oEnd, self.lAllObjects, self.oTokenMap)

    def get_tokens_matching(self, lTokens):
        return extract.get_tokens_matching(lTokens, self.lAllObjects, self.oTokenMap)

    def get_tokens_matching_not_at_beginning_or_ending_of_line(self, lTokens):
        return extract.get_tokens_matching_not_at_beginning_or_ending_of_line(lTokens, self.lAllObjects, self.oTokenMap)

    def get_n_token_after_tokens(self, iToken, lTokens):
        return extract.get_n_token_after_tokens(iToken, lTokens, self.lAllObjects, self.oTokenMap)

    def get_n_tokens_before_token(self, iN, lTokens):
        return extract.get_m_tokens_before_and_n_tokens_after_token(iN, 0, lTokens, self.lAllObjects, self.oTokenMap)

    def get_n_tokens_after_token(self, iN, lTokens):
        return extract.get_m_tokens_before_and_n_tokens_after_token(0, iN, lTokens, self.lAllObjects, self.oTokenMap)

    def get_m_tokens_before_and_n_tokens_after_token(self, iM, iN, lTokens):
        return extract.get_m_tokens_before_and_n_tokens_after_token(iM, iN, lTokens, self.lAllObjects, self.oTokenMap)

    def get_n_token_after_tokens_between_tokens(self, iToken, lTokens, oStart, oEnd):
        return extract.get_n_token_after_tokens_between_tokens(iToken, lTokens, oStart, oEnd, self.lAllObjects, self.oTokenMap)

    def get_tokens_matching_in_range_bounded_by_tokens(self, lTokens, oStart, oEnd):
        return extract.get_tokens_matching_in_range_bounded_by_tokens(lTokens, oStart, oEnd, self.lAllObjects, self.oTokenMap)

    def get_tokens_bounded_by(self, oLeft, oRight, include_trailing_whitespace=False, bExcludeLastToken=False, bIncludeTillEndOfLine=False, bIncludeTillBeginningOfLine=False):
        return extract.get_tokens_bounded_by(oLeft, oRight, self.lAllObjects, self.oTokenMap, include_trailing_whitespace=include_trailing_whitespace, bExcludeLastToken=bExcludeLastToken, bIncludeTillEndOfLine=bIncludeTillEndOfLine, bIncludeTillBeginningOfLine=bIncludeTillBeginningOfLine)

    def get_tokens_bounded_by_token_when_between_tokens(self, oLeft, oRight, oStart, oEnd, include_trailing_whitespace=False):
        return extract.get_tokens_bounded_by_token_when_between_tokens(oLeft, oRight, oStart, oEnd, self.lAllObjects, self.oTokenMap, include_trailing_whitespace)

    def get_tokens_bounded_by_tokens_if_token_is_between_them(self, oLeft, oRight, oToken):
        return extract.get_tokens_bounded_by_tokens_if_token_is_between_them(oLeft, oRight, oToken, self.lAllObjects, self.oTokenMap)

    def get_tokens_bounded_by_unless_between(self, oLeft, oRight, lUnless):
        return extract.get_tokens_bounded_by_unless_between(oLeft, oRight, lUnless, self.lAllObjects, self.oTokenMap)

    def get_tokens_at_beginning_of_line_matching(self, lTokens):
        return extract.get_tokens_at_beginning_of_line_matching(lTokens, self.lAllObjects, self.oTokenMap)

    def get_tokens_at_beginning_of_line_matching_unless_between_tokens(self, lTokens, lUnless):
        return extract.get_tokens_at_beginning_of_line_matching_unless_between_tokens(lTokens, lUnless, self.lAllObjects, self.oTokenMap)

    def get_tokens_at_beginning_of_line_matching_between_tokens(self, lTokens, oStart, oEnd, bInclusive=False):
        return extract.get_tokens_at_beginning_of_line_matching_between_tokens(lTokens, oStart, oEnd, bInclusive, self.lAllObjects, self.oTokenMap)

    def get_column_of_token_index(self, iToken):
        return extract.get_column_of_token_index(iToken, self.lAllObjects, self.oTokenMap)

    def get_line_above_line_starting_with_token(self, lTokens, bIncludeComments):
        return extract.get_line_above_line_starting_with_token(lTokens, self.lAllObjects, self.oTokenMap, bIncludeComments)

    def get_line_above_line_starting_with_token_with_hierarchy(self, lTokens, lHierarchy, bIncludeComments):
        return extract.get_line_above_line_starting_with_token_with_hierarchy(lTokens, self.lAllObjects, lHierarchy, self.oTokenMap, bIncludeComments)

    def get_line_below_line_ending_with_token(self, lTokens):
        return extract.get_line_below_line_ending_with_token(lTokens, self.lAllObjects, self.oTokenMap)

    def get_line_below_line_ending_with_several_possible_tokens(self, lTokens):
        return extract.get_line_below_line_ending_with_several_possible_tokens(lTokens, self.lAllObjects, self.oTokenMap)

    def get_line_below_line_ending_with_token_with_hierarchy(self, lTokens, lHierarchy):
        return extract.get_line_below_line_ending_with_token_with_hierarchy(lTokens, self.lAllObjects, lHierarchy, self.oTokenMap)

    def get_line_which_includes_tokens(self, lTokens):
        return extract.get_line_which_includes_tokens(lTokens, self.lAllObjects, self.oTokenMap)

    def get_if_statement_conditions(self, fRemoveWhitespace=True):
        return extract.get_if_statement_conditions(self.lAllObjects, self.oTokenMap, fRemoveWhitespace)

    def get_n_tokens_before_and_after_tokens(self, iToken, lTokens):
        return extract.get_n_tokens_before_and_after_tokens(iToken, lTokens, self.lAllObjects, self.oTokenMap)

    def get_n_tokens_before_and_after_tokens_bounded_by_tokens(self, iToken, lTokens, lBetween):
        return extract.get_n_tokens_before_and_after_tokens_bounded_by_tokens(iToken, lTokens, lBetween, self.lAllObjects, self.oTokenMap)

    def get_sequence_of_tokens_not_matching(self, lTokens):
        return extract.get_sequence_of_tokens_not_matching(lTokens, self.lAllObjects, self.oTokenMap)

    def get_tokens_between_tokens_inclusive_while_storing_value_from_token(self, left_token, right_token, value_token):
        return extract.get_tokens_between_tokens_inclusive_while_storing_value_from_token(left_token, right_token, value_token, self.lAllObjects, self.oTokenMap)

    def get_tokens_between_non_whitespace_token_and_token(self, right_token):
        return extract.get_tokens_between_non_whitespace_token_and_token(right_token, self.lAllObjects, self.oTokenMap)

    def get_tokens_from_line(self, iLineNumber):
        return extract.get_tokens_from_line(iLineNumber, self.lAllObjects, self.oTokenMap)

    def get_consecutive_lines_starting_with_token(self, search_token, min_num_lines=2):
        return extract.get_consecutive_lines_starting_with_token(search_token, min_num_lines, self.lAllObjects, self.oTokenMap)

    def get_consecutive_lines_starting_with_token_and_stopping_when_token_starting_line_is_found(self, search_token, stop_token):
        return extract.get_consecutive_lines_starting_with_token_and_stopping_when_token_starting_line_is_found(search_token, stop_token, self.lAllObjects, self.oTokenMap)

    def get_tokens_where_line_starts_with_token_until_ending_token_is_found(self, start_token, stop_token):
        return extract.get_tokens_where_line_starts_with_token_until_ending_token_is_found(start_token, stop_token, self.lAllObjects, self.oTokenMap)

    def get_token_and_n_tokens_before_it(self, oToken, iTokens):
        return extract.get_token_and_n_tokens_before_it(oToken, iTokens, self.lAllObjects, self.oTokenMap)

    def get_token_and_n_tokens_before_it_in_between_tokens(self, lTokens, iTokens, oStart, oEnd):
        return extract.get_token_and_n_tokens_before_it_in_between_tokens(lTokens, iTokens, oStart, oEnd, self.lAllObjects, self.oTokenMap)

    def get_token_and_n_tokens_before_it_in_between_tokens_unless_token_is_found(self, lTokens, iTokens, oStart, oEnd, oStop):
        return extract.get_token_and_n_tokens_before_it_in_between_tokens_unless_token_is_found(lTokens, iTokens, oStart, oEnd, oStop, self.lAllObjects, self.oTokenMap)

    def get_token_and_n_tokens_after_it(self, lTokens, iTokens):
        return extract.get_token_and_n_tokens_after_it(lTokens, iTokens, self.lAllObjects, self.oTokenMap)

    def get_token_and_n_tokens_after_it_when_between_tokens(self, lTokens, iTokens, oStart, oEnd):
        return extract.get_token_and_n_tokens_after_it_when_between_tokens(lTokens, iTokens, oStart, oEnd, self.lAllObjects, self.oTokenMap)

    def get_blank_lines_below_line_ending_with_token(self, lTokens, lHierarhcy=None):
        return extract.get_blank_lines_below_line_ending_with_token(lTokens, lHierarhcy, self.lAllObjects, self.oTokenMap)

    def get_blank_lines_below_line_ending_with_several_possible_tokens(self, lTokens):
        return extract.get_blank_lines_below_line_ending_with_several_possible_tokens(lTokens, self.lAllObjects, self.oTokenMap)

    def get_blank_lines_above_line_starting_with_token(self, lTokens):
        return extract.get_blank_lines_above_line_starting_with_token(lTokens, self.lAllObjects, self.oTokenMap)

    def get_blank_lines_above_line_starting_with_token_when_between_tokens(self, lTokens, lBetweenTokens):
        return extract.get_blank_lines_above_line_starting_with_token_when_between_tokens(lTokens, lBetweenTokens, self.lAllObjects, self.oTokenMap)

    def get_association_elements_between_tokens(self, oStart, oEnd):
        return extract.get_association_elements_between_tokens(oStart, oEnd, self.lAllObjects, self.oTokenMap)

    def get_interface_elements_between_tokens(self, oStart, oEnd, include_end_of_line_comments=False):
        return extract.get_interface_elements_between_tokens(oStart, oEnd, self.lAllObjects, self.oTokenMap)

    def get_lines_with_length_that_exceed_column(self, iColumn):
        return extract.get_lines_with_length_that_exceed_column(iColumn, self.lAllObjects, self.oTokenMap)

    def get_tokens_starting_with_token_and_ending_with_one_of_possible_tokens(self, lStartTokens, lEndTokens, bIncludeStartToken=False, bIncludeEndToken=True, bEarliestDetect=False):
        return extract.get_tokens_starting_with_token_and_ending_with_one_of_possible_tokens(lStartTokens, lEndTokens, self.lAllObjects, self.oTokenMap, bIncludeStartToken, bIncludeEndToken, bEarliestDetect)

    def get_tokens_between_indexes(self, iStartIndex, iEndIndex):
        return extract.get_tokens_between_indexes(iStartIndex, iEndIndex, self.lAllObjects)

    def get_tokens_from_beginning_of_line_containing_token_to_the_next_non_whitespace_token_to_the_right(self, token):
        return extract.get_tokens_from_beginning_of_line_containing_token_to_the_next_non_whitespace_token_to_the_right(token, self.lAllObjects, self.oTokenMap)

    def get_indent_of_line_at_index(self, iIndex):
        for iToken in range(iIndex, 0, -1):
            oToken = self.lAllObjects[iToken]
            if oToken.get_indent() is not None:
                return oToken.get_indent()
        return 0

    def get_tokens_in_declarative_parts(self):
        return extract.get_tokens_in_declarative_parts(self.lAllObjects, self.oTokenMap)


def split_on_carriage_return(lObjects):
    lReturn = []
    lMyObjects = []
    iLine = 1
    for oObject in lObjects:
        if isinstance(oObject, parser.carriage_return):
            lReturn.append(lMyObjects)
            iLine += 1
            lMyObjects = []
        else:
            lMyObjects.append(oObject)
    if len(lMyObjects) > 0:
        lReturn.append(lMyObjects)
    return lReturn


def post_token_assignments(lTokens):
    oCodeTags = code_tags.New()
    iParenId = 0
    lParenId = []
    for iToken, oToken in enumerate(lTokens):

        if isinstance(oToken, subtype_indication.type_mark) or isinstance(oToken, resolution_indication.resolution_function_name) or isinstance(oToken, type_mark.name):
            sValue = oToken.get_value()
            ### IEEE values
            if sValue.lower() == 'std_logic_vector':
                lTokens[iToken] = types.std_logic_vector(sValue)

            elif sValue.lower() == 'std_ulogic_vector':
                lTokens[iToken] = types.std_ulogic_vector(sValue)

            elif sValue.lower() == 'std_ulogic':
                lTokens[iToken] = types.std_ulogic(sValue)

            elif sValue.lower() == 'std_logic':
                lTokens[iToken] = types.std_logic(sValue)

            elif sValue.lower() == 'integer':
                lTokens[iToken] = types.integer(sValue)

            elif sValue.lower() == 'signed':
                lTokens[iToken] = types.signed(sValue)

        elif isinstance(oToken, parser.todo):
            sValue = oToken.get_value()
            if sValue == '&':
                lTokens[iToken] = adding_operator.concat()

            elif sValue  == '+':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = sign.plus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.keyword], iToken - 1, lTokens):
                    lTokens[iToken] = sign.plus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.assignment], iToken - 1, lTokens):
                    lTokens[iToken] = sign.plus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.comma], iToken - 1, lTokens):
                    lTokens[iToken] = sign.plus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([choices.bar], iToken - 1, lTokens):
                    lTokens[iToken] = sign.plus()
                else:
                    lTokens[iToken] = adding_operator.plus()

            elif sValue  == '-':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = sign.minus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.keyword], iToken - 1, lTokens):
                    lTokens[iToken] = sign.minus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.assignment], iToken - 1, lTokens):
                    lTokens[iToken] = sign.minus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.comma], iToken - 1, lTokens):
                    lTokens[iToken] = sign.minus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([choices.bar], iToken - 1, lTokens):
                    lTokens[iToken] = sign.minus()
                else:
                    lTokens[iToken] = adding_operator.minus()

            elif sValue == '(':
                lTokens[iToken] = parser.open_parenthesis()
                iParenId += 1
                lParenId.append(iParenId)
                lTokens[iToken].iId = iParenId

            elif sValue == ')':
                lTokens[iToken] = parser.close_parenthesis()
                lTokens[iToken].iId = lParenId.pop()

            elif sValue == ',':
                lTokens[iToken] = parser.comma()

            elif sValue.lower() == 'to':
                lTokens[iToken] = direction.to(sValue)

            elif sValue.lower() == 'downto':
                lTokens[iToken] = direction.downto(sValue)

            elif sValue.lower() == 'and':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.and_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.assignment], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.and_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.comma], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.and_operator(sValue)
                else:
                    lTokens[iToken] = logical_operator.and_operator(sValue)

            elif sValue.lower() == 'or':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.or_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.assignment], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.or_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.comma], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.or_operator(sValue)
                else:
                    lTokens[iToken] = logical_operator.or_operator(sValue)

            elif sValue.lower() == 'nand':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.nand_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.assignment], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.nand_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.comma], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.nand_operator(sValue)
                else:
                    lTokens[iToken] = logical_operator.nand_operator(sValue)

            elif sValue.lower() == 'nor':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.nor_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.assignment], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.nor_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.comma], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.nor_operator(sValue)
                else:
                    lTokens[iToken] = logical_operator.nor_operator(sValue)

            elif sValue.lower() == 'xor':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.xor_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.assignment], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.xor_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.comma], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.xor_operator(sValue)
                else:
                    lTokens[iToken] = logical_operator.xor_operator(sValue)

            elif sValue.lower() == 'xnor':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.xnor_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.assignment], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.xnor_operator(sValue)
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.comma], iToken - 1, lTokens):
                    lTokens[iToken] = unary_logical_operator.xnor_operator(sValue)
                else:
                    lTokens[iToken] = logical_operator.xnor_operator(sValue)

            elif sValue.lower() == '**':
                lTokens[iToken] = miscellaneous_operator.double_star(sValue)

            elif sValue.lower() == 'abs':
                lTokens[iToken] = miscellaneous_operator.abs_operator(sValue)

            elif sValue.lower() == 'not':
                lTokens[iToken] = miscellaneous_operator.not_operator(sValue)

            elif sValue.lower() == '*':
                lTokens[iToken] = multiplying_operator.star(sValue)

            elif sValue.lower() == '/':
                lTokens[iToken] = multiplying_operator.slash(sValue)

            elif sValue.lower() == 'mod':
                lTokens[iToken] = multiplying_operator.mod_operator(sValue)

            elif sValue.lower() == 'rem':
                lTokens[iToken] = multiplying_operator.rem_operator(sValue)

            elif sValue == '=':
                lTokens[iToken] = relational_operator.equal(sValue)

            elif sValue == "'":
                lTokens[iToken] = parser.tic(sValue)
                utils.classify_predefined_types(lTokens, iToken + 1)

            elif sValue.lower() == 'event':
                lTokens[iToken] = parser.event_keyword(sValue)

            ### IEEE values
            elif sValue.lower() == 'rising_edge':
                lTokens[iToken] = function.rising_edge(sValue)

            elif sValue.lower() == 'falling_edge':
                lTokens[iToken] = function.falling_edge(sValue)

            elif sValue.lower() == 'std_logic_vector':
                lTokens[iToken] = types.std_logic_vector(sValue)

            elif sValue.lower() == 'std_ulogic_vector':
                lTokens[iToken] = types.std_ulogic_vector(sValue)

            elif sValue.lower() == 'std_ulogic':
                lTokens[iToken] = types.std_ulogic(sValue)

            elif len(sValue) == 3 and sValue.startswith("'") and sValue.endswith("'"):
                lTokens[iToken] = parser.character_literal(sValue)
        else:
            sValue = oToken.get_value()
            if sValue  == '+':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = sign.plus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([exponent.e_keyword], iToken - 1, lTokens):
                    pass
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.keyword], iToken - 1, lTokens):
                    lTokens[iToken] = sign.plus()
                else:
                    lTokens[iToken] = adding_operator.plus()
            elif sValue  == '-':
                if utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis], iToken - 1, lTokens):
                    lTokens[iToken] = sign.minus()
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([exponent.e_keyword], iToken - 1, lTokens):
                    pass
                elif utils.are_previous_consecutive_token_types_ignoring_whitespace([parser.keyword], iToken - 1, lTokens):
                    lTokens[iToken] = sign.minus()
                else:
                    lTokens[iToken] = adding_operator.minus()
            elif sValue.lower() == '*':
                lTokens[iToken] = multiplying_operator.star(sValue)
            elif sValue.lower() == '/':
                lTokens[iToken] = multiplying_operator.slash(sValue)
            elif sValue.lower() == '**':
                lTokens[iToken] = miscellaneous_operator.double_star(sValue)
            elif sValue == '(':
                iParenId += 1
                lParenId.append(iParenId)
                oToken.iId = iParenId
            elif sValue == ')':
                oToken.iId = lParenId.pop()


def set_token_hierarchy_value(lTokens):
    iIfHierarchy = 0
    for oToken in lTokens:
        if isinstance(oToken, token.if_statement.if_keyword):
            oToken.set_hierarchy(iIfHierarchy)
            iIfHierarchy += 1
        if isinstance(oToken, token.if_statement.elsif_keyword):
            oToken.set_hierarchy(iIfHierarchy - 1)
        if isinstance(oToken, token.if_statement.else_keyword):
            oToken.set_hierarchy(iIfHierarchy - 1)
        if isinstance(oToken, token.if_statement.semicolon):
            iIfHierarchy -= 1
            oToken.set_hierarchy(iIfHierarchy)


def set_code_tags(lTokens):
    oCodeTags = code_tags.New()
    for oToken in lTokens:
        if code_tags.token_has_vsg_on_code_tag(oToken):
            oToken.set_code_tags(oCodeTags.get_tags())
            oCodeTags.update(oToken)
        elif code_tags.token_has_vsg_off_code_tag(oToken):
            oCodeTags.update(oToken)
            oToken.set_code_tags(oCodeTags.get_tags())
        elif code_tags.token_has_next_line_code_tag(oToken):
            oCodeTags.update(oToken)
            oToken.set_code_tags(oCodeTags.get_tags())
        else:
            oToken.set_code_tags(oCodeTags.get_tags())
            oCodeTags.update(oToken)


def set_aggregate_tokens(lTokens):
    lOpenParens = []
    for iToken, oToken in enumerate(lTokens):
        if isinstance(oToken, parser.open_parenthesis):
            lOpenParens.append(iToken)
        if isinstance(oToken, parser.close_parenthesis):
            iIndex = lOpenParens.pop()
            if isinstance(lTokens[iIndex], token.aggregate.open_parenthesis):
                iId = oToken.iId
                lTokens[iToken] = token.aggregate.close_parenthesis()
                lTokens[iToken].iId = iId
        if isinstance(oToken, token.element_association.assignment):
            iId = lTokens[lOpenParens[-1]].iId
            lTokens[lOpenParens[-1]] = token.aggregate.open_parenthesis()
            lTokens[lOpenParens[-1]].iId = iId


def combine_use_clause_selected_name(lTokens):
    lReturn = []
    for iToken, oToken in enumerate(lTokens):
        if isinstance(oToken, token.use_clause.selected_name):
            if isinstance(lReturn[-1], token.use_clause.selected_name):
                lReturn[-1].value = lReturn[-1].value + oToken.value
                continue
        lReturn.append(oToken)
    return lReturn


def remove_beginning_of_file_tokens(lTokens):
    lReturn = []
    for oToken in lTokens:
        if not isinstance(oToken, parser.beginning_of_file):
            lReturn.append(oToken)
    return lReturn


class options():

    def __init__(self):
        self.bInsideDelimitedComment = False

    def set_inside_delimited_comment(self):
        self.bInsideDelimitedComment = True

    def clear_inside_delimited_comment(self):
        self.bInsideDelimitedComment = False

    def inside_delimited_comment(self):
        return self.bInsideDelimitedComment




from .rule_201 import rule_201
from .rule_202 import rule_202
from .rule_203 import rule_203
from .rule_204 import rule_204
from .rule_205 import rule_205

from .rule_400 import rule_400
from .rule_401 import rule_401




from vsg import parser
from vsg import token
from vsg import violation

from vsg.vhdlFile import utils
from vsg.rule_group import structure

oInsertTokens = token.for_generate_statement.end_generate_label
oAnchorTokens = token.for_generate_statement.semicolon
oLeftTokens = token.for_generate_statement.end_keyword
oRightTokens = token.for_generate_statement.semicolon
oValueTokens = token.for_generate_statement.generate_label


class rule_011(structure.Rule):
    '''
    This rule checks the **end generate** line has a label on for generate statements.

    **Violation**

    .. code-block:: vhdl

       ram_array : for i in 0 to 127 generate

       end generate;

    **Fix**

    .. code-block:: vhdl

       ram_array : for i in 0 to 127 generate

       end generate ram_array;
    '''

    def __init__(self):
        structure.Rule.__init__(self, 'generate', '011')
        self.solution = 'Add generate label'
        self.insert_token = oInsertTokens
        self.anchor_token = oAnchorTokens
        self.left_token = oLeftTokens
        self.right_token = oRightTokens
        self.value_token = oValueTokens
        self.groups.append('structure::optional')
        self.configuration_documentation_link = None

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(token.architecture_body.begin_keyword, token.architecture_body.end_keyword)

    def _analyze(self, lToi):
        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)
            lLabels = []
            for iToken, oToken in enumerate(lTokens):
               iLine = utils.increment_line_number(iLine, oToken)

               if manage_labels(oToken, lLabels):
                   continue

               if isinstance(oToken, token.for_generate_statement.end_generate_keyword):
                   if not utils.are_next_consecutive_token_types_ignoring_whitespace([token.for_generate_statement.end_generate_label], iToken + 1, lTokens):
                       oNewToi = oToi.extract_tokens(iToken, iToken)
                       dAction = {}
                       dAction['label'] = token.for_generate_statement.end_generate_label(lLabels[-1].get_value())
                       sSolution = 'Add label ' + lLabels[-1].get_value()
                       oViolation = violation.New(oNewToi.get_line_number(), oNewToi, sSolution)
                       oViolation.set_action(dAction)
                       self.add_violation(oViolation)
                   continue

               if isinstance(oToken, token.if_generate_statement.end_generate_keyword):
                   if not utils.are_next_consecutive_token_types_ignoring_whitespace([token.if_generate_statement.end_generate_label], iToken + 1, lTokens):
                       oNewToi = oToi.extract_tokens(iToken, iToken)
                       dAction = {}
                       dAction['label'] = token.if_generate_statement.end_generate_label(lLabels[-1].get_value())
                       sSolution = 'Add label ' + lLabels[-1].get_value()
                       oViolation = violation.New(oNewToi.get_line_number(), oNewToi, sSolution)
                       oViolation.set_action(dAction)
                       self.add_violation(oViolation)
                   continue

               if isinstance(oToken, token.case_generate_statement.end_generate_keyword):
                   if not utils.are_next_consecutive_token_types_ignoring_whitespace([token.case_generate_statement.end_generate_label], iToken + 1, lTokens):
                       oNewToi = oToi.extract_tokens(iToken, iToken)
                       dAction = {}
                       dAction['label'] = token.case_generate_statement.end_generate_label(lLabels[-1].get_value())
                       sSolution = 'Add label ' + lLabels[-1].get_value()
                       oViolation = violation.New(oNewToi.get_line_number(), oNewToi, sSolution)
                       oViolation.set_action(dAction)
                       self.add_violation(oViolation)
                   continue

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        lTokens.append(parser.whitespace(' '))
        lTokens.append(dAction['label'])
        oViolation.set_tokens(lTokens)


def manage_labels(oToken, lLabels):

    if isinstance(oToken, token.for_generate_statement.generate_label):
        lLabels.append(oToken)
        return True

    if isinstance(oToken, token.if_generate_statement.generate_label):
        lLabels.append(oToken)
        return True

    if isinstance(oToken, token.case_generate_statement.generate_label):
        lLabels.append(oToken)
        return True

    if isinstance(oToken, token.for_generate_statement.semicolon):
        lLabels.pop()
        return True

    if isinstance(oToken, token.if_generate_statement.semicolon):
        lLabels.pop()
        return True

    if isinstance(oToken, token.case_generate_statement.semicolon):
        lLabels.pop()
        return True

    return False




from vsg.rules import blank_line_above_line_starting_with_token

from vsg import token

lTokens = []
lTokens.append(token.subprogram_body.end_keyword)


class rule_204(blank_line_above_line_starting_with_token):
    '''
    This rule checks for blank lines above the **end** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       begin

         a <= b;
       end function overflow;

    **Fix**

    .. code-block:: vhdl

       begin

         a <= b;

       end function overflow;
    '''

    def __init__(self):
        blank_line_above_line_starting_with_token.__init__(self, 'subprogram_body', '204', lTokens)




from vsg.token import architecture_body as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import architecture_declarative_part
from vsg.vhdlFile.classify import architecture_statement_part


def detect(iToken, lObjects):
    '''
    architecture identifier of *entity*_name is
        architecture_declarative_part
    begin
        architecture_statement_part
    end [ architecture ] [ *architecture*_simple_name ] ;
    '''

    if utils.is_next_token('architecture', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = classify_opening_declaration(iToken, lObjects)

    iCurrent = architecture_declarative_part.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('begin', token.begin_keyword, iCurrent, lObjects)

    iCurrent = architecture_statement_part.classify_until(['end'], iCurrent, lObjects)

    iCurrent = classify_closing_declaration(iToken, lObjects)

    return iCurrent


def classify_opening_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('architecture', token.architecture_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token(token.identifier, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('of', token.of_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token(token.entity_name, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)

    return iCurrent


def classify_closing_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if('architecture', token.end_architecture_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.architecture_simple_name, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent





from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure
from vsg.vhdlFile import utils


class move_token_next_to_another_token(structure.Rule):
    '''
    Moves one token next to another and places a single space between them.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    anchor_token : token type
       The token which another token will moved next to.

    token_to_move : token type
       The token which will be moved next to the anchor token.
    '''

    def __init__(self, name, identifier, anchor_token, token_to_move):
        structure.Rule.__init__(self, name, identifier)
        self.subphase = 2
        self.anchor_token = anchor_token
        self.token_to_move = token_to_move
        self.configuration_documentation_link = None

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(self.anchor_token, self.token_to_move, bIncludeTillEndOfLine=True)

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if not utils.are_next_consecutive_token_types([parser.whitespace, self.token_to_move], 1, lTokens) and \
               not utils.are_next_consecutive_token_types([self.token_to_move], 1, lTokens):
                oViolation = violation.New(oToi.get_line_number(), oToi, self.solution)
                oViolation.set_remap()
                oViolation.fix_blank_lines = True
                self.add_violation(oViolation)


    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        iIndex = oViolation.get_token_value()

        rules_utils.insert_token(lTokens, 1, lTokens.pop(iIndex))
        rules_utils.insert_whitespace(lTokens, 1)
        lNewTokens = utils.remove_consecutive_whitespace_tokens(lTokens)
        lNewTokens = utils.fix_blank_lines(lNewTokens)

        oViolation.set_tokens(lNewTokens)




from vsg import parser


class transport_keyword(parser.keyword):
    '''
    unique_id = delay_mechanism : transport_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class inertial_keyword(parser.keyword):
    '''
    unique_id = delay_mechanism : inertial_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class reject_keyword(parser.keyword):
    '''
    unique_id = delay_mechanism : reject_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)




from vsg.rules import align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens

from vsg import parser
from vsg import token

lAlign = []
lAlign.append(parser.comment)

lSkip = []
lSkip.append(parser.comment)


class rule_027(align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens):
    '''
    This rule checks the alignment of inline comments in the architecture declarative part.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       architecture rtl of my_entity is

         signal   wr_en    : std_logic;  -- Comment 1
         signal   rd_en    : std_logic;     -- Comment 2
         constant c_period : time; -- Comment 3

       begin

    **Fix**

    .. code-block:: vhdl

       architecture rtl of my_entity is

         signal   wr_en    : std_logic; -- Comment 1
         signal   rd_en    : std_logic; -- Comment 2
         constant c_period : time;      -- Comment 3

       begin
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens.__init__(self, 'architecture', '027', lAlign, token.architecture_body.is_keyword, token.architecture_body.begin_keyword, lSkip)
        self.solution = 'Align comments.'
        self.subphase = 4




from vsg import token

from vsg.rules.whitespace_before_token import Rule

lTokens = []
lTokens.append(token.entity_specification.colon)


class rule_101(Rule):
    '''
    This rule checks for at least a single space before the colon.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       attribute coordinate of comp_1: component is (0.0, 17.5);

       attribute coordinate of comp_1     : component is (0.0, 17.5);

    **Fix**

    .. code-block:: vhdl

       attribute coordinate of comp_1 : component is (0.0, 17.5);

       attribute coordinate of comp_1     : component is (0.0, 17.5);
    '''
    def __init__(self):
        Rule.__init__(self, 'entity_specification', '101', lTokens)
        self.number_of_spaces = '>=1'




from vsg.rules import token_case_with_prefix_suffix as Rule

from vsg import token

lTokens = []
lTokens.append(token.instantiated_unit.entity_name)


class rule_028(Rule):
    '''
    This rule checks the entity name has proper case in direct instantiations.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       instance_name : entity library.ENTITY_NAME

    **Fix**

    .. code-block:: vhdl

       instance_name : entity library.entity_name
    '''

    def __init__(self):
        Rule.__init__(self, 'instantiation', '028', lTokens)
        self.groups.append('case::name')




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.constant_declaration.constant_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of a constant declaration.

    **Violation**

    .. code-block:: vhdl

       architecture RTL of FIFO is

       constant size : integer := 1;
           constant width : integer := 32

    **Fix**

    .. code-block:: vhdl

       architecture RTL of FIFO is

         constant size : integer := 1;
         constant width : integer := 32
    '''

    def __init__(self):
        token_indent.__init__(self, 'constant', '001', lTokens)




from vsg.rules import align_tokens_in_region_between_tokens_unless_between_tokens

from vsg import token

lAlign = []
lAlign.append(token.full_type_declaration.identifier)
lAlign.append(token.incomplete_type_declaration.identifier)
lAlign.append(token.file_declaration.identifier)
lAlign.append(token.signal_declaration.identifier)
lAlign.append(token.constant_declaration.identifier)
lAlign.append(token.subtype_declaration.identifier)
lAlign.append(token.variable_declaration.identifier)
lAlign.append(token.alias_declaration.alias_designator)

oStart = token.process_statement.process_keyword
oEnd = token.process_statement.begin_keyword

lUnless = []
lUnless.append([token.subprogram_body.is_keyword,token.subprogram_body.begin_keyword])


class rule_031(align_tokens_in_region_between_tokens_unless_between_tokens):
    '''
    This rule checks for alignment of identifiers in the process declarative region.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       proc_1 : process(all) is

        variable     var1 : boolean;
        constant  cons1 : integer;
        file            file1  : load_file_file open read_mode is load_file_name;

       begin

       end process proc_1;

    **Fix**

    .. code-block:: vhdl

       proc_1 : process(all) is

        variable var1 : boolean;
        constant cons1 : integer;
        file     file1  : load_file_file open read_mode is load_file_name;

       begin

       end process proc_1;
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens_unless_between_tokens.__init__(self, 'process', '031', lAlign, oStart, oEnd, lUnless)
        self.solution = 'Align the first character of each identifier.'




from vsg.rules import blank_line_above_line_starting_with_token

from vsg.token import context_declaration as token


class rule_024(blank_line_above_line_starting_with_token):
    '''
    This rule checks for blank lines or comments above the **end** keyword.

    |configuring_blank_lines_link|

    The default style is :code:`no_code`.

    **Violation**

    .. code-block:: vhdl

         use ieee.std_logic_1164.all;
       end context;

    **Fix**

    .. code-block:: vhdl

         use ieee.std_logic_1164.all;

       end context;
    '''

    def __init__(self):
        blank_line_above_line_starting_with_token.__init__(self, 'context', '024', [token.end_keyword])




from vsg.token import null_statement as token

from vsg.vhdlFile import utils


def detect(iToken, lObjects):
    '''
    null_statement ::=
        [ label : ] null ;
    '''
    if utils.are_next_consecutive_tokens([None, ':', 'null'], iToken, lObjects):
        return classify(iToken, lObjects)
    if utils.is_next_token('null', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.label, token.label_colon)
    iCurrent = utils.assign_next_token_required('null', token.null_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent





from vsg.vhdlFile.extract import utils

from vsg.vhdlFile.extract.get_line_succeeding_line import get_line_succeeding_line


def get_line_below_line_ending_with_several_possible_tokens(lTokens, lAllTokens, oTokenMap):

    lReturn = []

    lTokenIndexes = utils.get_indexes_of_tokens_between(lTokens[0], lTokens[1:], oTokenMap)

    lIndexes = []
    for iIndex in lTokenIndexes:
        if utils.is_token_at_end_of_line(iIndex, oTokenMap):
            lIndexes.append(iIndex)

    lLine = utils.get_line_numbers_of_indexes_in_list(lIndexes, oTokenMap)
    for iLine in lLine:
        oToi = get_line_succeeding_line(iLine, lAllTokens, 1, oTokenMap)
        if oToi is not None:
            lReturn.append(oToi)

    return lReturn




from vsg import parser

from vsg.rules.whitespace_between_tokens import Rule as WhitespaceRule


class Rule(WhitespaceRule):
    '''
    Checks for a single space before a token.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    oFirstToken : token object
       Token to check for a single space before.

    oSecondToken : token object
       Token to use as an exclusion.
    '''

    def __init__(self, name, identifier):
        WhitespaceRule.__init__(self, name=name, identifier=identifier)
        self.lTokens = []
        self.oFirstToken = None
        self.oExceptToken = None

    def _get_tokens_of_interest(self, oFile):
        lFilter = self._create_filter_list(oFile)
        lToi = self._extract_tois(oFile)
        lReturn = filter_toi(lFilter, lToi)
        return lReturn

    def _extract_tois(self, oFile):
        lReturn = []
        lToi = oFile.get_token_and_n_tokens_before_it([self.oFirstToken], 2)
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if isinstance(lTokens[1], parser.whitespace):
                lReturn.append(oToi)
            else:
                lReturn.append(oToi.extract_tokens(1, 2))
        return lReturn

    def _create_filter_list(self, oFile):
        lExcept = oFile.get_line_which_includes_tokens([self.oExceptToken])
        lFilter = []
        for oExcept in lExcept:
            lFilter.append(oExcept.get_line_number())
        return lFilter


def filter_toi(lFilter, lToi):
    lReturn = []
    for oToi in lToi:
        if oToi.get_line_number() in lFilter:
            lReturn.append(oToi)
    return lReturn




from vsg import parser

from vsg.vhdlFile import utils

from vsg.vhdlFile.extract import tokens


def get_lines_with_length_that_exceed_column(iColumn, lAllTokens, oTokenMap):
    iLine = 1
    lReturn = []
    lTemp = []
    iStart = None
    bFirstTokenInLine = False
    for iToken, oToken in enumerate(lAllTokens):

        if isinstance(oToken, parser.carriage_return):
            if utils.does_length_of_tokens_exceed(lTemp, iColumn):
                lReturn.append(tokens.New(iStart, iLine, lTemp))
            iLine +=1
            lTemp = []
            bFirstTokenInLine = True
            continue

        lTemp.append(oToken)

        if bFirstTokenInLine:
            iStart = iToken

    return lReturn




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.block_statement.end_keyword, token.block_statement.end_block_keyword])
lTokens.append([token.block_statement.end_block_keyword, token.block_statement.end_block_label])


class rule_101(Rule):
    '''
    This rule checks for a single space between the **end** and **block** keywords and label.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end   block   block_label;

    **Fix**

    .. code-block:: vhdl

       end block block_label;
    '''
    def __init__(self):
        Rule.__init__(self, 'block', '101', lTokens)




from vsg.rules import separate_multiple_signal_identifiers_into_individual_statements

iAllow = 2


class rule_015(separate_multiple_signal_identifiers_into_individual_statements):
    '''
    This rule checks for multiple signal names defined in a single signal declaration.
    By default, this rule will only flag more than two signal declarations.

    |configuring_number_of_signals_in_signal_declaration_link|

    **Violation**

    .. code-block:: vhdl

       signal sig1, sig2
         sig3, sig4,
         sig5
         : std_logic;

    **Fix**

    .. code-block:: vhdl

       signal sig1 : std_logic;
       signal sig2 : std_logic;
       signal sig3 : std_logic;
       signal sig4 : std_logic;
       signal sig5 : std_logic;
    '''

    def __init__(self):
        separate_multiple_signal_identifiers_into_individual_statements.__init__(self, 'signal', '015', iAllow)




from vsg.rules import move_token_next_to_another_token

from vsg.token import context_declaration as token


class rule_005(move_token_next_to_another_token):
    '''
    This rule checks the context identifier is on the same line as the **context** keyword.

    **Violation**

    .. code-block:: vhdl

       context
       c1
         is

    **Fix**

    .. code-block:: vhdl

       context c1
         is
    '''

    def __init__(self):
        move_token_next_to_another_token.__init__(self, 'context', '005', token.context_keyword, token.identifier)
        self.subphase = 1
        self.solution = 'Move identifier next to *context* keyword'




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.full_type_declaration.is_keyword, token.enumeration_type_definition.open_parenthesis])
lTokens.append([token.full_type_declaration.is_keyword, token.range_constraint.range_keyword])
lTokens.append([token.full_type_declaration.is_keyword, token.unbounded_array_definition.array_keyword])
lTokens.append([token.full_type_declaration.is_keyword, token.constrained_array_definition.array_keyword])
lTokens.append([token.full_type_declaration.is_keyword, token.record_type_definition.record_keyword])
lTokens.append([token.full_type_declaration.is_keyword, token.access_type_definition.access_keyword])
lTokens.append([token.full_type_declaration.is_keyword, token.file_type_definition.file_keyword])


class rule_007(Rule):
    '''
    This rule checks for a single space after the **is** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       type state_machine is     (idle, write, read, done);

    **Fix**

    .. code-block:: vhdl

       type state_machine is (idle, write, read, done);
    '''
    def __init__(self):
        Rule.__init__(self, 'type', '007', lTokens)




from vsg import parser
from vsg import violation

from vsg.vhdlFile import utils
from vsg.rule_group import structure
from vsg.rules import utils as rules_utils


class insert_token_next_to_token_if_it_does_not_exist_between_tokens_using_value_from_token(structure.Rule):
    '''
    Checks for the existence of a token and will insert it if it does not exist.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    insert_token : token object
       token to insert if it does not exist.

    anchor_token : token object
       token to check if insert_token exists to the right of

    value_token : token object
       token to pull the value from
    '''

    def __init__(self, name, identifier):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.insert_token = None
        self.anchor_token = None
        self.left_token = None
        self.right_token = None
        self.value_token = None
        self.direction = None
        self.action = 'add'
        self.configuration.append('action')
        self.configuration_documentation_link = 'configuring_optional_items_link'
        self.filter_tokens = []

    def _get_tokens_of_interest(self, oFile):
        if remove_keyword(self):
            return oFile.get_token_and_n_tokens_before_it([self.insert_token], 1)
        else:
            lToi = oFile.get_tokens_between_tokens_inclusive_while_storing_value_from_token(self.left_token, self.right_token, self.value_token)
            return filter_toi(self.filter_tokens, lToi)

    def _analyze(self, lToi):
        if remove_keyword(self):
            analyze_for_existence_of_optional_keyword(lToi, self)
        else:
            analyze_for_missing_optional_keyword(lToi, self)

    def _fix_violation(self, oViolation):
        if remove_keyword(self):
            rules_utils.remove_optional_item(oViolation, self.insert_token)
        else:
            add_optional_item(oViolation, self)


def filter_toi(filter_tokens, lToi):
    lReturn = []
    for oToi in lToi:
        for oToken in filter_tokens:
            if oToi.token_type_exists(oToken):
                break
        else:
            lReturn.append(oToi)
    return lReturn


def remove_keyword(self):
    if self.action == 'remove':
        return True
    return False


def analyze_for_existence_of_optional_keyword(lToi, self):
    for oToi in lToi:
        oViolation = create_violation(oToi, oToi.get_line_number(), self)
        self.add_violation(oViolation)


def analyze_for_missing_optional_keyword(lToi, self):
    for oToi in lToi:
        iLine, lTokens = utils.get_toi_parameters(oToi)
        if not optional_keyword_exists(self.insert_token, lTokens):
            iLine += rules_utils.get_number_of_carriage_returns_before_token(self.anchor_token, lTokens)
            oViolation = create_violation(oToi, iLine, self)
            self.add_violation(oViolation)


def optional_keyword_exists(oToken, lTokens):
    if rules_utils.get_index_of_token_in_list(oToken, lTokens) is None:
        return False
    return True


def add_optional_item(oViolation, self):
    lTokens = oViolation.get_tokens()
    if not token_value_available(oViolation):
        return

    if not optional_keyword_exists(self.anchor_token, lTokens):
        return

    iIndex = rules_utils.get_index_of_token_in_list(self.anchor_token, lTokens)

    if self.direction == 'right':
        rules_utils.insert_token(lTokens, iIndex + 1, self.insert_token(oViolation.get_token_value()))
        rules_utils.insert_whitespace(lTokens, iIndex + 1)
    else:
        rules_utils.insert_token(lTokens, iIndex, self.insert_token(oViolation.get_token_value()))
        if not rules_utils.whitespace_before_token_index(lTokens, iIndex):
            rules_utils.insert_whitespace(lTokens, iIndex)

    oViolation.set_tokens(lTokens)


def token_value_available(oViolation):
    if oViolation.get_token_value() is not None:
        return True
    return False


def create_violation(oToi, iLineNumber, self):
    sSolution = self.action.capitalize() + ' ' + self.solution
    oViolation = violation.New(iLineNumber, oToi, sSolution)
    return oViolation




from vsg.vhdlFile import utils

from vsg.token import subtype_indication as token

from vsg.vhdlFile.classify import constraint
from vsg.vhdlFile.classify import resolution_indication
from vsg.vhdlFile.classify import type_mark


def classify(iToken, lObjects):
    '''
    subtype_indication ::=
        [ resolution_indication ] type_mark [ constraint ]
    '''

    iCurrent = resolution_indication.detect(iToken, lObjects)
    iCurrent = utils.find_next_non_whitespace_token(iCurrent, lObjects)
    iCurrent = type_mark.classify(iCurrent, lObjects)
    iCurrent = constraint.detect(iCurrent, lObjects)
    return iCurrent




from vsg.rules import remove_tokens_bounded_by_tokens_and_remove_trailing_whitespace

from vsg.token import case_statement as token


class rule_019(remove_tokens_bounded_by_tokens_and_remove_trailing_whitespace):
    '''
    This rule checks for labels before the **case** keyword.
    The label should be removed.
    The preference is to have comments above the case statement.

    **Violation**

    .. code-block:: vhdl

          CASE_LABEL : case address is
          CASE_LABEL: case address is
          case address is

    **Fix**

    .. code-block:: vhdl

          case address is
          case address is
          case address is
    '''

    def __init__(self):
        remove_tokens_bounded_by_tokens_and_remove_trailing_whitespace.__init__(self, 'case', '019', token.case_label, token.label_colon)
        self.solution = 'Remove Label'




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.concurrent_signal_assignment_statement.label_name)
lTokens.append(token.concurrent_signal_assignment_statement.postponed_keyword)
lTokens.append(token.concurrent_simple_signal_assignment.target)
lTokens.append(token.concurrent_conditional_signal_assignment.target)
lTokens.append(token.concurrent_selected_signal_assignment.with_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of concurrent assignments.

    **Violation**

    .. code-block:: vhdl

       architecture RTL of FIFO is
       begin

            wr_en <= '0';
       rd_en <= '1';

    **Fix**

    .. code-block:: vhdl

       architecture RTL of FIFO is
       begin

         wr_en <= '0';
         rd_en <= '1';
    '''

    def __init__(self):
        token_indent.__init__(self, 'concurrent', '001', lTokens)




from vsg import token

from vsg.rules.whitespace_after_token import Rule

lTokens = []
lTokens.append(token.report_statement.report_keyword)


class rule_100(Rule):
    '''
    This rule checks for a single space after the **report** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

        report     "FIFO width is limited to 16 bits.";

    **Fix**

    .. code-block:: vhdl

        report "FIFO width is limited to 16 bits.";
    '''
    def __init__(self):
        Rule.__init__(self, 'report_statement', '100', lTokens)




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.concurrent_procedure_call_statement.label_name, token.concurrent_procedure_call_statement.label_colon])
lTokens.append([token.concurrent_procedure_call_statement.label_colon, token.concurrent_procedure_call_statement.postponed_keyword])
lTokens.append([token.concurrent_procedure_call_statement.label_colon, token.procedure_call.procedure_name])
lTokens.append([token.concurrent_procedure_call_statement.postponed_keyword, token.procedure_call.procedure_name])

lTokens.append([token.procedure_call_statement.label, token.procedure_call_statement.label_colon])
lTokens.append([token.procedure_call_statement.label_colon, token.procedure_call.procedure_name])


class rule_100(Rule):
    '''
    This rule checks for a single space between the following block elements:  label, label colon, **postponed** keyword and the *procedure* name.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       procedure_label   :    postponed   WR_EN(parameter);

    **Fix**

    .. code-block:: vhdl

       procedure_label : postponed WR_EN(parameter);
    '''
    def __init__(self):
        Rule.__init__(self, 'procedure_call', '100', lTokens)
        self.solution = 'Ensure a single space between the label, colon, *postponed* keyword and procedure_name.'




from vsg import parser

from vsg.token import sensitivity_list as token

from vsg.vhdlFile import utils


def classify(iToken, lObjects):
    '''
    subtype_indication ::=
        [ resolution_indication ] type_mark [ constraint ]
    '''

    iCurrent = iToken
    iStop = len(lObjects) - 1
    iOpenParenthesis = 0
    iCloseParenthesis = 0
    while iCurrent < iStop:
        iCurrent = utils.find_next_token(iCurrent, lObjects)
        if utils.token_is_open_parenthesis(iCurrent, lObjects):
           iOpenParenthesis += 1
        if utils.token_is_close_parenthesis(iCurrent, lObjects):
           iCloseParenthesis += 1
        if iOpenParenthesis < iCloseParenthesis:
            break
        else:
            if utils.is_next_token(',', iCurrent, lObjects):
                utils.assign_token(lObjects, iCurrent, token.comma)
            else:
                utils.assign_token(lObjects, iCurrent, parser.todo)
    return iCurrent


def classify_until(lUntils, iToken, lObjects):
    '''
    sensitivity_list ::=
        *signal*_name { , *signal*_name}
    '''

    iCurrent = iToken
    iLast = 0
    while iLast != iCurrent:
        iLast = iCurrent
        if lObjects[utils.find_next_token(iCurrent, lObjects)].get_value().lower() in lUntils:
            return iCurrent
        iCurrent = utils.assign_next_token_if(',', token.comma, iCurrent, lObjects)
        iCurrent = utils.assign_next_token(parser.todo, iCurrent, lObjects)




from vsg.rules import multiline_constraint_structure as Rule

from vsg import token

lTokenPairs = []
lTokenPairs.append([token.variable_declaration.variable_keyword, token.variable_declaration.semicolon])


class rule_017(Rule):
    '''
    This rule checks the structure of variable constraints.

    |configuring_multiline_constraint_rules_link|

    .. NOTE:: The indenting of multiline variable constraints is handled by the rule `variable_400 <variable_rules.html#variable-400>`_.

    **Violation**

    .. code-block:: vhdl

       variable v_element : my_record(element1(7 downto 0), element2(3 downto 0));

    **Fix**

    .. code-block:: vhdl

       variable v_element : my_record(
           element1(7 downto 0),
           element2(3 downto 0)
         );
    '''

    def __init__(self):
        Rule.__init__(self, 'variable', '017')
        self.lTokenPairs = lTokenPairs




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.case_generate_statement.generate_label)
lTokens.append(token.for_generate_statement.generate_label)
lTokens.append(token.if_generate_statement.generate_label)


class rule_001(token_indent):
    '''
    This rule checks the indent of the generate declaration.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is
       begin

       ram_array : for i in 0 to 7 generate

             ram_array : for i in 0 to 7 generate

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is
       begin

         ram_array : for i in 0 to 7 generate

         ram_array : for i in 0 to 7 generate
    '''

    def __init__(self):
        token_indent.__init__(self, 'generate', '001', lTokens)




from vsg.rules import align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens

from vsg import parser
from vsg import token

lAlign = []
lAlign.append(parser.comment)

lSkip = []
lSkip.append(parser.comment)

class rule_020(align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens):
    '''
    This rule checks for alignment of inline comments in the entity declaration.

    Following extra configurations are supported:

    * :code:`separate_generic_port_alignment`.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       generic (
           g_width        : positive;  -- Data width
           g_output_delay : positive -- Delay at output
       );
       port (
           clk_i  : in std_logic; -- Input clock
           data_i : in std_logic;   -- Data input
           data_o : in std_logic -- Data output
       );

    **Fix**

    .. code-block:: vhdl

       generic (
           g_width        : positive; -- Data width
           g_output_delay : positive  -- Delay at output
       );
       port (
           clk_i  : in std_logic; -- Input clock
           data_i : in std_logic; -- Data input
           data_o : in std_logic  -- Data output
       );
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens_skipping_lines_starting_with_tokens.__init__(self, 'entity', '020', lAlign, token.entity_declaration.entity_keyword, token.entity_declaration.end_keyword, lSkip)
        self.solution = 'Align comment.'
        self.subphase = 4




from vsg import parser


class multiplying_operator(parser.keyword):
    '''
    unique_id = multiplying_operator : multiplying_operator
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class star(multiplying_operator):
    '''
    unique_id = multiplying_operator : star
    '''

    def __init__(self, sString='*'):
        multiplying_operator.__init__(self, '*')


class slash(multiplying_operator):
    '''
    unique_id = multiplying_operator : slash
    '''

    def __init__(self, sString='/'):
        multiplying_operator.__init__(self, '/')


class mod_operator(multiplying_operator):
    '''
    unique_id = multiplying_operator : mod_operator
    '''

    def __init__(self, sString):
        multiplying_operator.__init__(self, sString)


class rem_operator(multiplying_operator):
    '''
    unique_id = multiplying_operator : rem_operator
    '''

    def __init__(self, sString):
        multiplying_operator.__init__(self, sString)




from vsg import rules


class token_case_with_prefix_suffix(rules.token_case):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of token types
       The token types to check the case on.
    '''

    def __init__(self, name, identifier, lTokens):
        rules.token_case.__init__(self, name=name, identifier=identifier, lTokens=lTokens)
        self.configuration.append('prefix_exceptions')
        self.configuration.append('suffix_exceptions')
        self.configuration.append('case_exceptions')




from vsg.token import assertion
from vsg.token import assertion_statement
from vsg.token import concurrent_assertion_statement

from vsg.rules import align_consecutive_lines_after_line_starting_with_token_and_stopping_with_token as Rule


class rule_400(Rule):
    '''
    This rule checks the alignment of the report expressions.

    .. NOTE:: There is a configuration option **alignment** which changes the indent location of multiple lines.

    **Violation**

    .. code-block:: vhdl

       assert WIDTH > 16
         report "FIFO width is limited" &
       " to 16 bits."
         severity FAILURE;

    **Fix**

    .. code-block:: vhdl

       assert WIDTH > 16
         report "FIFO width is limited" &
                " to 16 bits."
         severity FAILURE;
    '''

    def __init__(self):
        Rule.__init__(self, name="assert", identifier="400")
        self.lStartTokens = [assertion.report_keyword]
        self.lEndTokens = [assertion.severity_keyword, assertion_statement.semicolon, concurrent_assertion_statement.semicolon]
        self.configuration_documentation_link = None




from vsg.rules.whitespace_after_token import Rule

from vsg.token import iteration_scheme as token

lTokens = []
lTokens.append(token.for_keyword)


class rule_101(Rule):
    '''
    This rule checks that a single space exists after the **for** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       for      x in (31 downto 0) loop

    **Fix**

    .. code-block:: vhdl

       for x in (31 downto 0) loop
    '''
    def __init__(self):
        Rule.__init__(self, 'iteration_scheme', '101', lTokens)
        self.solution = 'Ensure a single space after for keyword.'




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.entity_declaration.entity_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of the **entity** keyword.

    **Violation**

    .. code-block:: vhdl

       library ieee;

         entity fifo is

    **Fix**

    .. code-block:: vhdl

       library ieee;

       entity fifo is
    '''

    def __init__(self):
        token_indent.__init__(self, 'entity', '001', lTokens)




from vsg.rule_group import case
from vsg.rules import case_utils
from vsg.rules import utils


class token_case(case.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of token types
       The token types to check the case on.
    '''

    def __init__(self, name, identifier, lTokens):
        case.Rule.__init__(self, name=name, identifier=identifier)
        self.solution = None
        self.phase = 6
        self.case = 'lower'
        self.configuration.append('case')
        self.lTokens = lTokens
        self.prefix_exceptions = []
        self.suffix_exceptions = []
        self.case_exceptions = []

    def _get_tokens_of_interest(self, oFile):
        self.case_exceptions_lower = utils.lowercase_list(self.case_exceptions)
        return oFile.get_tokens_matching(self.lTokens)

    def _analyze(self, lToi):
        check_prefix = case_utils.is_exception_enabled(self.prefix_exceptions)
        check_suffix = case_utils.is_exception_enabled(self.suffix_exceptions)
        check_whole = case_utils.is_exception_enabled(self.case_exceptions)
        for oToi in lToi:
            oViolation = case_utils.check_for_case_violation(oToi, self, check_prefix, check_suffix, check_whole)
            if oViolation is not None:
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        lTokens[0].set_value(dAction['value'])
        oViolation.set_tokens(lTokens)




from vsg.rules import token_indent_between_tokens

from vsg import token

lTokens = []
lTokens.append(token.component_instantiation_statement.instantiation_label)
lTokens.append(token.component_instantiation_statement.semicolon)

lTokens.append(token.generic_map_aspect.generic_keyword)
lTokens.append(token.generic_map_aspect.close_parenthesis)

lTokens.append(token.port_map_aspect.port_keyword)
lTokens.append(token.port_map_aspect.close_parenthesis)

lTokens.append(token.association_element.formal_part)

oStart = token.component_instantiation_statement.instantiation_label
oEnd = token.component_instantiation_statement.semicolon


class rule_001(token_indent_between_tokens):
    '''
    This rule checks for the proper indentation of instantiations.

    **Violation**

    .. code-block:: vhdl

         U_FIFO : FIFO
      port map (
               WR_EN    => wr_en,
       RD_EN    => rd_en,
             OVERFLOW => overflow
                    );

    **Fix**

    .. code-block:: vhdl

       U_FIFO : FIFO
         port map (
           WR_EN    => wr_en,
           RD_EN    => rd_en,
           OVERFLOW => overflow
         );
    '''

    def __init__(self):
        token_indent_between_tokens.__init__(self, 'instantiation', '001', lTokens, oStart, oEnd, bInclusive=True)




from vsg.rules import blank_line_below_line_ending_with_token

from vsg import token

lTokens = []
lTokens.append(token.process_statement.begin_keyword)


class rule_022(blank_line_below_line_ending_with_token):
    '''
    This rule checks for a blank line below the **begin** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is
       begin
         rd_en <= '0';

    **Fix**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is
       begin

         rd_en <= '0';
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'process', '022', lTokens)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.generic_clause.close_parenthesis)


class rule_008(token_indent):
    '''
    This rule checks the indent of the closing parenthesis.

    **Violation**

    .. code-block:: vhdl

       g_depth : integer := 512
       );

    **Fix**

    .. code-block:: vhdl

         g_depth : integer := 512
       );
    '''

    def __init__(self):
        token_indent.__init__(self, 'generic', '008', lTokens)




from vsg.rules import move_token_next_to_another_token as Rule

from vsg.token import entity_declaration as token


class rule_022(Rule):
    '''
    This rule checks the identifier is on the same line as the **entity** keyword.

    **Violation**

    .. code-block:: vhdl

       entity fifo is

       entity
         fifo is

    **Fix**

    .. code-block:: vhdl

       entity fifo is

       entity fifo is
    '''

    def __init__(self):
        Rule.__init__(self, 'entity', '022', token.entity_keyword, token.identifier)
        self.subphase = 1
        self.solution = 'Move identifier next to *entity* keyword'




from vsg.rules import align_tokens_in_region_between_tokens

from vsg import token

lAlign = []
lAlign.append(token.interface_unknown_declaration.colon)


class rule_017(align_tokens_in_region_between_tokens):
    '''
    This rule checks the alignment of the colon for each generic and port in the entity declaration.

    Following extra configurations are supported:

    * :code:`separate_generic_port_alignment`.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       generic (
           g_width : positive;
           g_output_delay : positive
       );
       port (
           clk_i : in std_logic;
           data_i : in std_logic;
           data_o : in std_logic
       );

    **Fix**

    .. code-block:: vhdl

       generic (
           g_width        : positive;
           g_output_delay : positive
       );
       port (
           clk_i  : in std_logic;
           data_i : in std_logic;
           data_o : in std_logic
       );
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens.__init__(self, 'entity', '017', lAlign, token.entity_declaration.entity_keyword, token.entity_declaration.end_keyword)
        self.solution = 'Align :.'




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.block_statement.block_label, token.block_statement.label_colon])
lTokens.append([token.block_statement.label_colon, token.block_statement.block_keyword])
lTokens.append([token.block_statement.block_keyword, token.block_statement.is_keyword])
lTokens.append([token.block_statement.block_keyword, token.block_statement.guard_open_parenthesis])
lTokens.append([token.block_statement.guard_close_parenthesis, token.block_statement.is_keyword])


class rule_100(Rule):
    '''
    This rule checks for a single space between the following block elements:  label, label colon, **block** keyword, guard open parenthesis, guart close parenthesis, and **is** keywords.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       block_label    :    block    (guard_condition)   is
       block_label  :   block    is

    **Fix**

    .. code-block:: vhdl

       block_label : block (guard_condition) is
       block_label : block is
    '''
    def __init__(self):
        Rule.__init__(self, 'block', '100', lTokens)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.null_statement.label)
lTokens.append(token.null_statement.null_keyword)


class rule_013(token_indent):
    '''
    This rule checks the indent of the **null** keyword.

    **Violation**

    .. code-block:: vhdl

         when others =>
            null;

         when others =>
       null;

    **Fix**

    .. code-block:: vhdl

       when others =>
         null;

       when others =>
         null;
    '''

    def __init__(self):
        token_indent.__init__(self, 'case', '013', lTokens)




from vsg.vhdlFile import utils


def classify_until(lUntils, iToken, lObjects):

    iCurrent = iToken
    iLast = 0
    while iLast != iCurrent:
        iLast = iCurrent
        if lObjects[utils.find_next_token(iCurrent, lObjects)].get_value().lower() in lUntils:
            return iCurrent
        iCurrent = utils.assign_next_token(parser.todo, iCurrent, lObjects)




from vsg.rules import utils as rules_utils
from vsg.vhdlFile import utils

from vsg import violation


def check_for_carriage_return_after_token(self, oToi):
    if self.value == 'ignore':
        return []
    elif self.value == 'no':
        return check_no(self, oToi)
    elif self.value == 'yes':
        return check_yes(self, oToi)


def check_no(self, oToi):    
    return check(self, oToi, 'remove_new_line')


def check_yes(self, oToi):    
    return check(self, oToi, 'add_new_line')


def check(self, oToi, action):
    lTokens = oToi.get_tokens()
    iLine = oToi.get_line_number()
    lReturn = []

    value = set_check_value(action)

    for iToken, oToken in enumerate(lTokens):
        if rules_utils.token_exists_in_token_type_list(oToken, self.analysis_options):
            if utils.is_token_at_end_of_line(iToken, lTokens) == value:
                iViolation_line_number = iLine + rules_utils.number_of_carriage_returns(lTokens[0:iToken])
                sSolution = 'jcl-fix this'
                iEndIndex = utils.find_next_non_whitespace_token(iToken + 1, lTokens) - 1
                oMyToi = oToi.extract_tokens(iToken + 1, iEndIndex)
                oViolation = violation.New(iViolation_line_number, oMyToi, sSolution)
                dAction = {}
                dAction['action'] = action
                oViolation.set_action(dAction)
                lReturn.append(oViolation)
    return lReturn


def set_check_value(action):
    if action == 'remove_new_line':
        return True
    return False


def check_for_carriage_return_before_token(self, oToi):
    if self.value == 'ignore':
        return []
    lTokens = oToi.get_tokens()
    iLine = oToi.get_line_number()
    lReturn = []

    for iToken, oToken in enumerate(lTokens):
        if rules_utils.token_exists_in_token_type_list(oToken, self.analysis_options):
            if rules_utils.token_at_beginning_of_line_in_token_list(iToken, lTokens) and self.value == 'no':
                sSolution = 'jcl-fix this'
                iViolation_line_number = iLine + rules_utils.number_of_carriage_returns(lTokens[0:iToken])
                iStartIndex = utils.find_previous_non_whitespace_token(iToken - 1, lTokens) + 1
                oMyToi = oToi.extract_tokens(iStartIndex, iToken - 1)
                oViolation = violation.New(iViolation_line_number, oMyToi, 'jcl-fix this')
                dAction = {}
                dAction['action'] = 'remove_new_line'
                oViolation.set_action(dAction)
                lReturn.append(oViolation)
            elif not rules_utils.token_at_beginning_of_line_in_token_list(iToken, lTokens) and self.value == 'yes':
                sSolution = 'jcl-fix this'
                iViolation_line_number = iLine + rules_utils.number_of_carriage_returns(lTokens[0:iToken])
                iStartIndex = utils.find_previous_non_whitespace_token(iToken - 1, lTokens) + 1
                oMyToi = oToi.extract_tokens(iStartIndex, iToken - 1)
                oViolation = violation.New(iViolation_line_number, oMyToi, 'jcl-fix this')
                dAction = {}
                dAction['action'] = 'add_new_line'
                oViolation.set_action(dAction)
                lReturn.append(oViolation)
    return lReturn


def check_for_carriage_returns_between_tokens_ignoring_leading_and_trailing_whitespace(self, oToi):
    if self.value == 'ignore':
        return []
    lTokens = oToi.get_tokens()
    iLine = oToi.get_line_number()
    lReturn = []
    for lTokenPairs in self.analysis_options:
        for iToken, oToken in enumerate(lTokens):
            if isinstance(oToken, lTokenPairs[0]):
                iEndIndex = rules_utils.get_index_of_token_in_list(lTokenPairs[1], lTokens)
                if iEndIndex is None:
                    continue
#                print(f'{iToken}, {iEndIndex}')
                iStartIndex = utils.find_next_non_whitespace_token(iToken + 1, lTokens)
                iEndIndex = utils.find_previous_non_whitespace_token(iEndIndex - 1, lTokens)
                iNumberCarriageReturns = rules_utils.number_of_carriage_returns(lTokens[iStartIndex:iEndIndex + 1])
#                print(f'{iNumberCarriageReturns}')
                if iNumberCarriageReturns > 0:
                    sSolution = 'jcl-fix this'

#                    print(lTokens[iStartIndex:iEndIndex + 1])
                    iViolation_line_number = iLine + rules_utils.number_of_carriage_returns(lTokens[0:iStartIndex])
                    oMyToi = oToi.extract_tokens(iStartIndex, iEndIndex)
#                    print(oMyToi.get_tokens())
                    oViolation = violation.New(iViolation_line_number, oMyToi, 'jcl-fix this')
                    dAction = {}
                    dAction['action'] = 'remove_new_line'
                    oViolation.set_action(dAction)
                    lReturn.append(oViolation)
    return lReturn




import copy

from vsg import parser
from vsg import violation

from vsg.token import port_clause as token
from vsg.token import interface_unknown_declaration
from vsg.token import interface_list
from vsg.rule_group import structure


class rule_026(structure.Rule):
    '''
    This rule checks for multiple identifiers on port declarations.

    Any comments are not replicated.

    **Violation**

    .. code-block:: vhdl

       port (
         wr_en, rd_en : in    std_logic;  -- Comment
         data     : inout std_logic;
         overflow, empty : out   std_logic -- Other comment
       );

    **Fix**

    .. code-block:: vhdl

       port (
         wr_en    : in    std_logic;
         rd_en    : in    std_logic;  -- Comment
         data    : inout std_logic
         overflow : out   std_logic;
         empty : out   std_logic -- Other comment
       );
    '''

    def __init__(self):
        structure.Rule.__init__(self, name='port', identifier='026')
        self.subphase = 2
        self.configuration_documentation_link = None

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_interface_elements_between_tokens(token.open_parenthesis, token.close_parenthesis)

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iCount = 0
            lIdentifiers = []
            lIdentifierIndexes = []
            for iToken, oToken in enumerate(lTokens):
                if isinstance(oToken, interface_unknown_declaration.identifier):
                    lIdentifiers.append(oToken.get_value())
                    iCount += 1
                    lIdentifierIndexes.append(iToken)
            if iCount > 1:
                sSolution = 'Split identifiers ' + ', '.join(lIdentifiers) + ' to individual lines.'
                dAction = {}
                if oToi == lToi[-1]:
                    dAction['last_element'] = False
                else:
                    dAction['last_element'] = True
                dAction['identifier_indexes'] = lIdentifierIndexes
                dAction['split_index'] = lIdentifierIndexes[-1] + 1
                oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
                oViolation.set_action(dAction)
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        '''
        Applies fixes for any rule violations.
        '''
        lTokens = oViolation.get_tokens()
        lNewTokens = []
        dAction = oViolation.get_action()
        for iIndex in dAction['identifier_indexes']:
            lCopyTokens = duplicate_tokens(lTokens)
            lNewTokens.append(lCopyTokens[iIndex])

            lNewTokens.extend(lCopyTokens[dAction['split_index']:])
            if iIndex != dAction['identifier_indexes'][-1]:
                lNewTokens.append(interface_list.semicolon())
                lNewTokens.append(parser.carriage_return())
        oViolation.set_tokens(lNewTokens)


def duplicate_tokens(lTokens):
    lReturn = []
    for oToken in lTokens:
        lReturn.append(copy.deepcopy(oToken))
    return lReturn




from vsg.rules import insert_token_right_of_token_if_it_does_not_exist_before_token

from vsg.token import architecture_body as token


class rule_010(insert_token_right_of_token_if_it_does_not_exist_before_token):
    '''
    This rule checks for the keyword **architecture** in the **end architecture** statement.
    It is clearer to the reader to state what is ending.

    |configuring_optional_items_link|

    **Violation**

    .. code-block:: vhdl

       end architecture_name;

    **Fix**

    .. code-block:: vhdl

       end architecture architecture_name;
    '''
    def __init__(self):
        insert_token_right_of_token_if_it_does_not_exist_before_token.__init__(self, 'architecture', '010', token.end_architecture_keyword('architecture'), token.end_keyword, token.semicolon)
        self.solution = 'architecture keyword.'
        self.groups.append('structure::optional')




from vsg.token import protected_type_body as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import protected_type_body_declarative_part


def detect(iToken, lObjects):
    '''
    protected_type_body ::=
        **protected** **body**
        protected_type_body_declarative_part
        **end** **protected** **body** [ protected_type_simple_name ]
    '''

    if utils.are_next_consecutive_tokens(['protected', 'body'], iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('protected', token.protected_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('body', token.body_keyword, iToken, lObjects)

    iCurrent = protected_type_body_declarative_part.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('protected', token.end_protected_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('body', token.end_body_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.protected_type_simple_name, iCurrent, lObjects)

    return iCurrent




from vsg.rules import align_tokens_in_region_between_tokens

from vsg import token

lAlign = []
lAlign.append(token.association_element.assignment)

oStart = token.component_instantiation_statement.label_colon
oEnd = token.component_instantiation_statement.semicolon


class rule_010(align_tokens_in_region_between_tokens):
    '''
    This rule checks the alignment of the **=>** operator for each generic and port in the instantiation.

    Following extra configurations are supported:

    * :code:`separate_generic_port_alignment`.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       U_FIFO : FIFO
         generic map (
           g_width => 8,
           g_delay    => 2
         )
         port map (
           wr_en => wr_en,
           rd_en => rd_en,
           overflow => overflow
         );

    **Fix**

    .. code-block:: vhdl

       U_FIFO : FIFO
         generic map (
           g_width => 8,
           g_delay => 2
         )
         port map (
           wr_en    => wr_en,
           rd_en    => rd_en,
           overflow => overflow
         );
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens.__init__(self, 'instantiation', '010', lAlign, oStart, oEnd)
        self.solution = 'Inconsistent alignment of "=>" in generic or port assignments of instantiation.'




from vsg import token

from vsg.rules.whitespace_before_token import Rule

lTokens = []
lTokens.append(token.simple_force_assignment.assignment)
lTokens.append(token.simple_waveform_assignment.assignment)
lTokens.append(token.simple_release_assignment.assignment)


class rule_003(Rule):
    '''
    This rule checks for at least a single space before the **<=** operator.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en<= '1';
       rd_en   <= '0';

    **Fix**

    .. code-block:: vhdl

       wr_en <= '1';
       rd_en   <= '0';
    '''
    def __init__(self):
        Rule.__init__(self, 'sequential', '003', lTokens)
        self.number_of_spaces = '>=1'




from vsg.token import protected_type_body as token

from vsg.vhdlFile.extract.get_tokens_bounded_by import get_tokens_bounded_by

oStart = token.body_keyword
oEnd = token.end_keyword


def extract(lAllTokens, oTokenMap):

    lReturn = []
    iStart = 1
    lToi = get_tokens_bounded_by(oStart, oEnd, lAllTokens, oTokenMap)
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        iEnd = len(lTokens) - 1
        lReturn.append(oToi.extract_tokens(iStart, iEnd))
    return lReturn




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.subprogram_body.end_keyword, token.subprogram_kind.function_keyword])
lTokens.append([token.subprogram_body.end_keyword, token.subprogram_body.designator])
lTokens.append([token.subprogram_kind.function_keyword, token.subprogram_body.designator])


class rule_101(Rule):
    '''
    This rule checks for a single space between the **end** and **function** keywords and function designator.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end   function   overflow;
       end   function;
       end   overflow;

    **Fix**

    .. code-block:: vhdl

       end function overflow;
       end function;
       end overflow;
    '''
    def __init__(self):
        Rule.__init__(self, 'function', '101', lTokens)




from vsg import parser


class procedure_name(parser.item):
    '''
    unique_id = procedure_call : procedure_name
    '''

    def __init__(self, sString):
        parser.item.__init__(self, sString)


class open_parenthesis(parser.open_parenthesis):
    '''
    unique_id = procedure_call : open_parenthesis
    '''

    def __init__(self, sString='('):
        parser.open_parenthesis.__init__(self)


class close_parenthesis(parser.close_parenthesis):
    '''
    unique_id = procedure_call : close_parenthesis
    '''

    def __init__(self, sString=')'):
        parser.close_parenthesis.__init__(self)




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.process_statement.is_keyword)


class rule_013(token_case):
    '''
    This rule checks the **is** keyword has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) IS
       begin

    **Fix**

    .. code-block:: vhdl

       proc_a : process (rd_en, wr_en, data_in, data_out,
                         rd_full, wr_full
                        ) is
       begin
    '''

    def __init__(self):
        token_case.__init__(self, 'process', '013', lTokens)
        self.groups.append('case::keyword')




from vsg.rule_group import indent
from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils


class token_indent(indent.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of token types
       token type to apply the indent rule
    '''

    def __init__(self, name, identifier, lTokens):
        indent.Rule.__init__(self, name=name, identifier=identifier)
        self.lTokens = lTokens

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_at_beginning_of_line_matching(self.lTokens)

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            if indent_should_be_zero_but_has_leading_whitespace(lTokens):
                create_zero_indent_violation(self, oToi)
            elif indent_exists_but_is_incorrect(self, lTokens):
                create_indent_violation(self, oToi, lTokens)
            elif no_indent_exists_but_should(self, lTokens):
                create_no_indent_violation(self, oToi, lTokens)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        if oViolation.get_action() == 'remove_whitespace':
            oViolation.set_tokens([lTokens[1]])
        elif oViolation.get_action() == 'adjust_whitespace':
            if self.indentStyle == 'spaces':
                lTokens[0].set_value(lTokens[1].get_indent() * self.indentSize * ' ')
                lTokens[0].has_tab = False
            elif self.indentStyle == 'smart_tabs':
                lTokens[0].set_value(lTokens[1].get_indent() * '\t')
                lTokens[0].has_tab = True
            oViolation.set_tokens(lTokens)
        elif oViolation.get_action() == 'add_whitespace':
            if self.indentStyle == 'spaces':
                rules_utils.insert_whitespace(lTokens, 0, lTokens[0].get_indent() * self.indentSize)
            else:
                rules_utils.insert_whitespace(lTokens, 0, lTokens[0].get_indent(), '\t')
            oViolation.set_tokens(lTokens)


def indent_should_be_zero_but_has_leading_whitespace(lTokens):
    if len(lTokens) == 2 and lTokens[1].get_indent() == 0:
        return True
    return False


def create_zero_indent_violation(self, oToi):
    sSolution = "Indent level 0"
    create_violation(self, oToi, sSolution, 'remove_whitespace')


def indent_exists_but_is_incorrect(self, lTokens):
    if len(lTokens) == 2:
        if lTokens[1].get_indent() is None:
            return False
        sWhitespace = lTokens[0].get_value()
        if self.indentStyle == 'spaces':
            sIndent = ' ' * (self.indentSize * lTokens[1].get_indent())
        else:
            sIndent = '\t' * (lTokens[1].get_indent())
        if sWhitespace != sIndent:
            return True
    return False


def create_indent_violation(self, oToi, lTokens):
    sSolution = create_adjust_indent_solution(self, lTokens)
    create_violation(self, oToi, sSolution, 'adjust_whitespace')


def no_indent_exists_but_should(self, lTokens):
    if not len(lTokens) == 1:
        return False

    if lTokens[0].get_indent() is None:
        return False
    if self.indentSize == 0:
        return False
    if lTokens[0].get_indent() != 0:
        return True

    return False


def create_no_indent_violation(self, oToi, lTokens):
    sSolution = create_no_indent_solution(self, lTokens)
    create_violation(self, oToi, sSolution, 'add_whitespace')


def create_violation(self, oToi, sSolution, sAction):
    oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
    oViolation.set_action(sAction)
    self.add_violation(oViolation)


def create_adjust_indent_solution(self, lTokens):
    return create_indent_solution(self, lTokens[1])


def create_no_indent_solution(self, lTokens):
    return create_indent_solution(self, lTokens[0])


def create_indent_solution(self, oToken):
    iIndentLevel = oToken.get_indent()
    if self.indentStyle == 'spaces':
        sSolution = 'Use ' + str(iIndentLevel * self.indentSize) + ' spaces for indent'
    elif self.indentStyle == 'smart_tabs':
        sSolution = 'Use ' + str(iIndentLevel) + ' tab(s) for indent'
    return sSolution




from vsg import violation

from vsg.rule_group import structure


class remove_tokens_bounded_by_tokens_and_remove_trailing_whitespace(structure.Rule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    left_token : token object
       The first token that defines the region

    right_token : token object
       The second token that defines the region
    '''

    def __init__(self, name, identifier, left_token, right_token):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.left_token = left_token
        self.right_token = right_token
        self.configuration_documentation_link = None

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(self.left_token, self.right_token, include_trailing_whitespace=True)

    def _analyze(self, lToi):
        for oToi in lToi:
           self.add_violation(violation.New(oToi.get_line_number(), oToi, self.solution))

    def _fix_violation(self, oViolation):
        oViolation.set_tokens([])




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.package_body.package_keyword, token.package_body.body_keyword])
lTokens.append([token.package_body.body_keyword, token.package_body.package_simple_name])
lTokens.append([token.package_body.package_simple_name, token.package_body.is_keyword])


class rule_100(Rule):
    '''
    This rule checks for a single space between **package**, **body** and **is** keywords.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       package    body  FIFO_PKG   is

    **Fix**

    .. code-block:: vhdl

       package body FIFO_PKG is
    '''
    def __init__(self):
        Rule.__init__(self, 'package_body', '100', lTokens)




from vsg.rules import token_prefix

from vsg import token

lTokens = []
lTokens.append(token.constant_declaration.identifier)


class rule_015(token_prefix):
    '''
    This rule checks for valid prefixes on constant identifiers.
    The default constant prefix is *c\_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       constant my_const : integer;

    **Fix**

    .. code-block:: vhdl

       constant c_my_const : integer;
    '''

    def __init__(self):
        token_prefix.__init__(self, 'constant', '015', lTokens)
        self.prefixes = ['c_']




from vsg.rules import previous_line

from vsg.token import component_declaration as token


class rule_003(previous_line):
    '''
    This rule checks for blank lines or comments above the **component** declaration.

    |configuring_previous_line_rules_link|

    The default style is :code:`no_code`.

    **Violation**

    .. code-block:: vhdl

       end component fifo;
       component ram is

    **Fix**

    .. code-block:: vhdl

       end component fifo;

       component ram is
    '''

    def __init__(self):
        previous_line.__init__(self, 'component', '003', [token.component_keyword])
        self.style = 'no_code'




from vsg.token import generic_map_aspect as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import association_list


def detect(iToken, lObjects):
    '''
    generic_map_aspect ::=
        generic map ( *generic*_association_list )
    '''
    if utils.is_next_token('generic', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_if('generic', token.generic_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if('map', token.map_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if('(', token.open_parenthesis, iCurrent, lObjects)

    iCurrent = association_list.classify(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_if(')', token.close_parenthesis, iCurrent, lObjects)

    return iCurrent




from vsg import parser


class target(parser.target):
    '''
    unique_id = simple_force_assignment : target
    '''

    def __init__(self, sString):
        parser.target.__init__(self, sString)


class assignment(parser.assignment):
    '''
    unique_id = simple_force_assignment : assignment
    '''

    def __init__(self, sString):
        parser.assignment.__init__(self, sString)


class force_keyword(parser.keyword):
    '''
    unique_id = simple_force_assignment : force_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = simple_force_assignment : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg.token import block_statement as token

from vsg.vhdlFile.extract.get_tokens_bounded_by import get_tokens_bounded_by

oStart = token.block_keyword
oEnd = token.begin_keyword


def extract(lAllTokens, oTokenMap):

    return get_tokens_bounded_by(oStart, oEnd, lAllTokens, oTokenMap)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.port_clause.port_keyword)


class rule_002(token_indent):
    '''
    This rule checks the indent of the **port** keyword.

    **Violation**

    .. code-block:: vhdl

       entity FIFO is
       port (

    **Fix**

    .. code-block:: vhdl

       entity FIFO is
         port (
    '''

    def __init__(self):
        token_indent.__init__(self, 'port', '002', lTokens)




from vsg.vhdlFile.extract import tokens
from vsg.vhdlFile.extract import utils


def get_token_and_n_tokens_before_it_in_between_tokens_unless_token_is_found(lTokens, iTokens, oStart, oEnd, oStop, lAllTokens, oTokenMap):
    lReturn = []

    lIndexes = utils.filter_tokens_between_tokens_unless_token_exists_between_them(lTokens, oStart, oEnd, oStop, oTokenMap)

    for iIndex in lIndexes:
        iStart = iIndex - iTokens
        if iStart >= 0:
            iLine = oTokenMap.get_line_number_of_index(iStart)
            lReturn.append(tokens.New(iStart, iLine, lAllTokens[iStart:iIndex + 1]))

    return lReturn




from vsg import parser
from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.simple_variable_assignment.assignment, parser.todo])
lTokens.append([token.conditional_variable_assignment.assignment, parser.todo])
lTokens.append([token.selected_variable_assignment.assignment, parser.todo])


class rule_002(Rule):
    '''
    This rule checks for a single space after the assignment.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

         counter :=0;
         count   :=     counter + 1;

    **Fix**

    .. code-block:: vhdl

         counter := 0;
         count   := counter + 1;
    '''
    def __init__(self):
        Rule.__init__(self, 'variable_assignment', '002', lTokens)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.generic_clause.generic_keyword)


class rule_002(token_indent):
    '''
    This rule checks the indent of the **generic** keyword.

    **Violation**

    .. code-block:: vhdl

       entity fifo is
            generic (

       entity fifo is
       generic (

    **Fix**

    .. code-block:: vhdl

       entity fifo is
         generic (

       entity fifo is
         generic (
    '''

    def __init__(self):
        token_indent.__init__(self, 'generic', '002', lTokens)




from vsg.rules import blank_line_below_line_ending_with_token

from vsg.token import loop_statement as token


class rule_203(blank_line_below_line_ending_with_token):
    '''
    This rule checks for blank lines below the **end loop** keywords.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       end loop;
       a <= b;

    **Fix**

    .. code-block:: vhdl

       end loop;

       a <= b;
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'loop_statement', '203', [token.semicolon])




from vsg import parser


class when_keyword(parser.keyword):
    '''
    unique_id = case_statement_alternative : when_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class assignment(parser.assignment):
    '''
    unique_id = case_statement_alternative : assignment
    '''

    def __init__(self, sString):
        parser.assignment.__init__(self, sString)


class others_keyword(parser.keyword):
    '''
    unique_id = case_statement_alternative : others_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)




from vsg import parser
from vsg.token import delimited_comment as token


def classify(lTokens, lObjects, oOptions):

    if len(lObjects) == 0 and oOptions.inside_delimited_comment():
        lObjects.append(token.text(''))

    for iToken, sToken in enumerate(lTokens):
        classify_delimited_comment_text(iToken, lObjects, oOptions)
        classify_opening_comment_delimiters(iToken, lObjects, oOptions)
        classify_closing_comment_delimiters(iToken, lObjects, oOptions)

    merge_text_tokens(lObjects)


def classify_closing_comment_delimiters(iToken, lObjects, oOptions):
    sToken = lObjects[iToken].get_value()
    if oOptions.inside_delimited_comment() and sToken == '*/':
        lObjects[iToken] = token.ending(sToken)
        oOptions.clear_inside_delimited_comment()


def classify_opening_comment_delimiters(iToken, lObjects, oOptions):
    classify_single_line_comment(iToken, lObjects, oOptions)
    classify_delimited_comment_open_keyword(iToken, lObjects, oOptions)


def classify_single_line_comment(iToken, lObjects, oOptions):
    sToken = lObjects[iToken].get_value()
    if not oOptions.inside_delimited_comment() and sToken.startswith('--'):
        lObjects[iToken] = parser.comment(sToken)
        if '\t' in sToken:
            lObjects[iToken].has_tab = True


def classify_delimited_comment_open_keyword(iToken, lObjects, oOptions):
    sToken = lObjects[iToken].get_value()
    if not oOptions.inside_delimited_comment() and sToken == '/*':
        lObjects[iToken] = token.beginning(sToken)
        oOptions.set_inside_delimited_comment()


def classify_delimited_comment_text(iToken, lObjects, oOptions):
    sToken = lObjects[iToken].get_value()
    if oOptions.inside_delimited_comment():
        lObjects[iToken] = token.text(sToken)
        if '\t' in sToken:
            lObjects[iToken].has_tab = True


def merge_text_tokens(lObjects):

    iStartIndex, iEndIndex = find_start_and_end_index_of_text_tokens(lObjects)
    merge_tokens(iStartIndex, iEndIndex, lObjects)


def find_start_and_end_index_of_text_tokens(lObjects):
    iStartIndex = -1
    iEndIndex = -1
    for iToken, oToken in enumerate(lObjects):
        iStartIndex = set_start_index(oToken, iToken, iStartIndex)
        iEndIndex = set_end_index(oToken, iToken, iEndIndex)
    return iStartIndex, iEndIndex


def set_start_index(oToken, iToken, iStartIndex):
    if isinstance(oToken, token.text) and iStartIndex == -1:
        iStartIndex = iToken
    return iStartIndex


def set_end_index(oToken, iToken, iEndIndex):
    if isinstance(oToken, token.text):
        iEndIndex = iToken
    return iEndIndex


def merge_tokens(iStartIndex, iEndIndex, lObjects):
    if iStartIndex < iEndIndex:
        sNewValue = ''
        for iIndex in range(iStartIndex, iEndIndex + 1):
            sNewValue += lObjects[iIndex].get_value()
        del lObjects[iStartIndex:iEndIndex + 1]
        lObjects.insert(iStartIndex, token.text(sNewValue))




from vsg import deprecated_rule


class rule_002(deprecated_rule.Rule):
    '''
    This rule has been split into the following rules:

    * :ref:`architecture_030`
    * :ref:`architecture_031`
    * :ref:`architecture_032`
    * :ref:`architecture_033`
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'architecture', '002')
        self.message.append('Rule ' + self.unique_id + ' has been split into the following rules:')
        self.message.append('  architecture_030')
        self.message.append('  architecture_031')
        self.message.append('  architecture_032')
        self.message.append('  architecture_033')




from vsg import parser
from vsg.token import block_statement as token

from vsg.rules import move_token_sequences_left_of_token

lSequences = []
lSequences.append([token.block_label, token.label_colon, token.block_keyword])
lSequences.append([token.block_label, token.label_colon, parser.whitespace, token.block_keyword])
lSequences.append([token.block_label, parser.whitespace, token.label_colon, token.block_keyword])
lSequences.append([token.block_label, parser.whitespace, token.label_colon, parser.whitespace, token.block_keyword])

oLeftToken = token.block_keyword


class rule_001(move_token_sequences_left_of_token):
    '''
    This rule checks the block label and the **block** keyword are on the same line.
    Keeping the label and generate on the same line reduces excessive indenting.

    **Violation**

    .. code-block:: vhdl

       block_label :
         block is

    **Fix**

    .. code-block:: vhdl

       block_label : block is
    '''

    def __init__(self):
        move_token_sequences_left_of_token.__init__(self, 'block', '001', lSequences, oLeftToken)
        self.solution = 'Ensure block label is on the same line as *block* keyword.'




from vsg import parser


class group_keyword(parser.keyword):
    '''
    unique_id = group_declaration : group_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class identifier(parser.identifier):
    '''
    unique_id = group_declaration : identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class colon(parser.colon):
    '''
    unique_id = group_declaration : colon
    '''

    def __init__(self, sString=':'):
        parser.colon.__init__(self)


class group_template_name(parser.name):
    '''
    unique_id = group_declaration : group_template_name
    '''

    def __init__(self, sString):
        parser.name.__init__(self, sString)


class open_parenthesis(parser.open_parenthesis):
    '''
    unique_id = group_specification : open_parenthesis
    '''

    def __init__(self, sString='('):
        parser.open_parenthesis.__init__(self)


class close_parenthesis(parser.close_parenthesis):
    '''
    unique_id = group_specification : close_parenthesis
    '''

    def __init__(self, sString=')'):
        parser.close_parenthesis.__init__(self)


class semicolon(parser.semicolon):
    '''
    unique_id = group_declaration : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens as Rule

from vsg import token

lTokens = []
lTokens.append(token.association_list.comma)

lTokenPairs = []
lTokenPairs.append([token.port_map_aspect.open_parenthesis, token.port_map_aspect.close_parenthesis])


class rule_009(Rule):
    '''
    This rule checks multiple port assignments on the same line.

    **Violation**

    .. code-block:: vhdl

       port map (
         WR_EN => w_wr_en, RD_EN => w_rd_en,
         OVERFLOW => w_overflow
       );

    **Fix**

    .. code-block:: vhdl

       port map (
         WR_EN => w_wr_en,
         RD_EN => w_rd_en,
         OVERFLOW => w_overflow
       );
    '''

    def __init__(self):
        Rule.__init__(self, 'port_map', '009', lTokens, lTokenPairs)
        self.solution = 'Move multiple port assignments to their own lines.'





from vsg.vhdlFile.classify import array_type_definition
from vsg.vhdlFile.classify import record_type_definition


def detect(iToken, lObjects):
    '''
    composite_type_definition ::=
        array_type_definition
      | record_type_definition
    '''

    iCurrent = array_type_definition.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = record_type_definition.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    return iToken
